# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Optional

import torch
import torch.nn as nn
import torch.nn.functional as F

_compiled_compute_cross_entropy = None


def compute_cross_entropy(
    logits: torch.Tensor,
    targets: torch.Tensor,
    ignore_index=-100,
    reduction="sum",
):
    """Computes the cross-entropy loss between logits and targets.

    Args:
        logits (torch.Tensor): Model predictions of shape (sequence_length, num_classes).
        targets (torch.Tensor): Ground-truth labels of shape (sequence_length,).
        ignore_index (int, optional): Target value that is ignored when computing the loss.
            Defaults to -100.

    Returns:
        torch.Tensor: The sum of cross-entropy losses over the sequence.
    """
    return F.cross_entropy(logits.float(), targets, ignore_index=ignore_index, reduction=reduction)


class ChunkedCrossEntropy(nn.Module):
    def __init__(self, chunk_len: int = 32, compile: bool = True, ignore_index: int = -100, reduction: str = "sum"):
        """
        Chunked cross-entropy loss.

        Args:
            chunk_len (int, optional): The size of each chunk. The sequence will be split
                along the first dimension in chunks of this length. Defaults to 32.
            compile (bool, optional): If True, uses the compiled compute_cross_entropy function.
                Defaults to True.
            ignore_index (int, optional): Target value that is ignored when computing the loss.
                Defaults to -100.
            reduction (str, optional): Type of reduction. Defaults to "sum".
        """
        super().__init__()
        self.chunk_len = chunk_len
        self.compile = compile
        self.ignore_index = ignore_index
        self.reduction = reduction

    def forward(
        self,
        logits: torch.Tensor,
        labels: torch.Tensor,
        mask: Optional[torch.Tensor] = None,
        num_label_tokens: Optional[int] = None,
    ) -> torch.Tensor:
        """Computes cross-entropy loss in chunks to handle long sequences more efficiently.

        Args:
            logits (torch.Tensor): Model output logits of shape [batch_size, seq_len, vocab_size].
            labels (torch.Tensor): Ground-truth labels of shape [batch_size, seq_len].
            mask (torch.Tensor, optional): Boolean mask indicating valid positions (1) and
                positions to ignore (0). Defaults to None.

        Returns:
            torch.Tensor: The sum of cross-entropy losses over the sequence.
        """
        # copied the following block from masked_ce
        # this may happen with CPUOffloadPolicy
        if labels.device != logits.device:
            labels = labels.to(logits.device)  # pragma: no cover
        # reshape to (N, C) and (N,) respectively
        logits = logits.view(-1, logits.size(-1))
        labels = labels.view(-1)
        if mask is not None:
            with torch.no_grad():
                if mask.device != labels.device:
                    mask = mask.to(labels.device)  # pragma: no cover
                labels.masked_fill_(mask.view(-1) == 0, self.ignore_index)
                del mask

        # maybe refactor if this is moved to a class?
        global _compiled_compute_cross_entropy
        if _compiled_compute_cross_entropy is None:
            _compiled_compute_cross_entropy = torch.compile(compute_cross_entropy, dynamic=True)

        seq_len = logits.shape[0]
        num_chunks = (seq_len + self.chunk_len - 1) // self.chunk_len
        loss = 0.0
        for logits_chunk, targets_chunk in zip(logits.chunk(num_chunks, dim=0), labels.chunk(num_chunks, dim=0)):
            loss += _compiled_compute_cross_entropy(logits_chunk, targets_chunk, self.ignore_index, self.reduction)
        if num_label_tokens is not None:
            assert self.reduction == "sum", "num_label_tokens is only supported when reduction is 'sum'"
            loss = loss / num_label_tokens  # pragma: no cover
        return loss



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from transformers import PretrainedConfig


class DeepseekV32Config(PretrainedConfig):
    model_type = "deepseek_v32"
    keys_to_ignore_at_inference = ["past_key_values"]

    def __init__(
        self,
        vocab_size: int = 129280,
        hidden_size: int = 7168,
        intermediate_size: int = 18432,
        moe_intermediate_size: int = 2048,
        num_hidden_layers: int = 61,
        num_attention_heads: int = 128,
        num_key_value_heads: int = 128,
        n_shared_experts: int = 1,
        n_routed_experts: int = 256,
        ep_size: int = 1,
        routed_scaling_factor: float = 2.5,
        kv_lora_rank: int = 512,
        q_lora_rank: int = 1536,
        qk_head_dim: int = 192,
        qk_nope_head_dim: int = 128,
        qk_rope_head_dim: int = 64,
        v_head_dim: int = 128,
        hidden_act: str = "silu",
        max_position_embeddings: int = 163840,
        initializer_range: float = 0.02,
        rms_norm_eps: float = 1e-6,
        use_cache: bool = True,
        pad_token_id: int | None = None,
        bos_token_id: int = 0,
        eos_token_id: int = 1,
        pretraining_tp: int = 1,
        tie_word_embeddings: bool = False,
        rope_theta: float = 10000.0,
        rope_scaling: dict | None = None,
        attention_bias: bool = False,
        attention_dropout: float = 0.0,
        num_experts_per_tok: int = 8,
        n_group: int = 8,
        topk_group: int = 4,
        first_k_dense_replace: int = 3,
        norm_topk_prob: bool = True,
        scoring_func: str = "sigmoid",
        aux_loss_alpha: float = 0.001,
        seq_aux: bool = True,
        # V3.2 Indexer parameters
        index_n_heads: int = 64,
        index_head_dim: int = 128,
        index_topk: int = 2048,
        torch_dtype: str = "bfloat16",
        **kwargs,
    ):
        self.vocab_size = vocab_size
        self.max_position_embeddings = max_position_embeddings
        self.hidden_size = hidden_size
        self.intermediate_size = intermediate_size
        self.moe_intermediate_size = moe_intermediate_size
        self.num_hidden_layers = num_hidden_layers
        self.num_attention_heads = num_attention_heads
        self.num_key_value_heads = num_key_value_heads
        self.n_shared_experts = n_shared_experts
        self.n_routed_experts = n_routed_experts
        self.ep_size = ep_size
        self.routed_scaling_factor = routed_scaling_factor
        self.kv_lora_rank = kv_lora_rank
        self.q_lora_rank = q_lora_rank
        self.qk_head_dim = qk_head_dim
        self.qk_nope_head_dim = qk_nope_head_dim
        self.qk_rope_head_dim = qk_rope_head_dim
        self.v_head_dim = v_head_dim
        self.hidden_act = hidden_act
        self.initializer_range = initializer_range
        self.rms_norm_eps = rms_norm_eps
        self.pretraining_tp = pretraining_tp
        self.use_cache = use_cache
        self.rope_theta = rope_theta
        self.rope_scaling = rope_scaling
        self.attention_bias = attention_bias
        self.attention_dropout = attention_dropout
        self.num_experts_per_tok = num_experts_per_tok
        self.n_group = n_group
        self.topk_group = topk_group
        self.first_k_dense_replace = first_k_dense_replace
        self.norm_topk_prob = norm_topk_prob
        self.scoring_func = scoring_func
        self.aux_loss_alpha = aux_loss_alpha
        self.seq_aux = seq_aux

        # V3.2 Indexer parameters
        self.index_n_heads = index_n_heads
        self.index_head_dim = index_head_dim
        self.index_topk = index_topk

        self.torch_dtype = torch_dtype

        super().__init__(
            pad_token_id=pad_token_id,
            bos_token_id=bos_token_id,
            eos_token_id=eos_token_id,
            tie_word_embeddings=tie_word_embeddings,
            **kwargs,
        )



# Copyright (c) 2025, NVIDIA CORPORATION. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""HuggingFace-compatible checkpointing mixin for NeMo Automodel.

This module provides a mixin class that gives models HuggingFace-compatible
save_pretrained() and from_pretrained() methods while using NeMo's checkpointing
infrastructure internally.

Key design principle: We do NOT override state_dict() or load_state_dict().
PyTorch's DCP expects these to behave like standard nn.Module methods.
HF format conversions happen only in save_pretrained() and from_pretrained() via Checkpointer.

Checkpointer is passed explicitly (dependency injection) - no global state.
"""

import logging
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from transformers.tokenization_utils import PreTrainedTokenizerBase

    from nemo_automodel.components.checkpoint.checkpointing import Checkpointer

logger = logging.getLogger(__name__)


class HFCheckpointingMixin:
    """Mixin providing HF-compatible API using NeMo's checkpointing infrastructure.

    Provides save_pretrained() and from_pretrained() methods that use Checkpointer
    for unified distributed/async support with HF format conversion.

    Key design: We do NOT override state_dict() or load_state_dict() because
    PyTorch's DCP expects these to behave like standard nn.Module methods.

    For PreTrainedModel subclasses:
    - super().from_pretrained() handles: downloads, quantization config, meta device init
    - Checkpointer.load_base_model() handles: actual weight loading with format conversion

    For nn.Module subclasses (no parent from_pretrained):
    - Falls back to manual config loading + Checkpointer
    """

    def save_pretrained(
        self,
        save_directory: str,
        checkpointer: Optional["Checkpointer"] = None,
        tokenizer: Optional["PreTrainedTokenizerBase"] = None,
        **kwargs,
    ) -> None:
        """Save model in HF-compatible format using Checkpointer infrastructure.

        Supports distributed saving, sharding, and async checkpointing.

        Args:
            save_directory: Output path
            checkpointer: Checkpointer instance. Uses self._checkpointer if not provided.
            tokenizer: Optional tokenizer to save alongside model
            **kwargs: Additional arguments
        """
        if checkpointer is None:
            raise ValueError("No checkpointer provided. Please pass the `checkpointer` argument.")

        # Use Checkpointer.save_model() which handles:
        # - ModelState.state_dict()
        # - _maybe_adapt_state_dict_to_hf() conversion
        # - Distributed/sharded saving via DCP
        # - Consolidated HF safetensors output
        # - Async checkpointing if enabled
        checkpointer.save_model(
            model=self,
            weights_path=save_directory,
            peft_config=kwargs.get("peft_config", None),
            tokenizer=tokenizer,
        )



from .fp8 import (
    HAVE_TORCHAO,
    FP8Config,
    apply_fp8_to_model,
    build_fp8_config,
    create_fp8_config_from_dict,
    verify_fp8_conversion,
)
from .qlora import (
    HAS_BNB,
    create_bnb_config,
    verify_qlora_quantization,
)

if HAVE_TORCHAO:
    from torchao.float8 import Float8LinearConfig
else:
    Float8LinearConfig = None

__all__ = [
    "apply_fp8_to_model",
    "verify_fp8_conversion",
    "build_fp8_config",
    "create_fp8_config_from_dict",
    "HAVE_TORCHAO",
    "FP8Config",
    "HAS_BNB",
    "create_bnb_config",
    "verify_qlora_quantization",
]

if HAVE_TORCHAO:
    __all__.append("Float8LinearConfig")



# Copyright (c) 2020, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
import logging
import os
from contextlib import contextmanager

from nemo_automodel.shared.import_utils import safe_import

HAVE_TORCHAO, torch_ao = safe_import("torchao")

import torch
import torch.nn as nn

logger = logging.getLogger(__name__)


def _supports_logits_to_keep(model: nn.Module) -> bool:
    """
    Check if the model supports logits_to_keep.

    Args:
        model (nn.Module): The model to check.

    Returns:
        bool: True if the model supports logits_to_keep, False otherwise.
    """
    if callable(getattr(model, "forward", None)):
        return "logits_to_keep" in set(inspect.signature(model.forward).parameters.keys())
    else:
        return False


def _supports_seq_lens(model: nn.Module) -> bool:
    """
    Check if the model's forward() accepts seq_lens.

    Returns True if:
    - forward() has an explicit `seq_lens` parameter, OR
    - forward() has **kwargs (so it won't crash if seq_lens is passed)

    Returns False otherwise (passing seq_lens would cause "unexpected kwarg" error).
    """
    if not callable(getattr(model, "forward", None)):
        return False
    try:
        sig = inspect.signature(model.forward)
        params = sig.parameters
        # Check for explicit seq_lens parameter
        if "seq_lens" in params:
            return True
        # Check for **kwargs (VAR_KEYWORD)
        for param in params.values():
            if param.kind == inspect.Parameter.VAR_KEYWORD:
                return True
        return False
    except (ValueError, TypeError):
        return False


def _get_model_param_stats(model: nn.Module) -> tuple[int, int, float]:
    """
    Get the number of trainable parameters and the L2 norm of the model.

    Args:
        model: Model to analyze

    Returns:
        total_params: int
        trainable_params: int
        local_sq_norm: float
    """
    total_params = 0
    trainable_params = 0
    local_sq_norm = 0.0

    for p in model.parameters():
        n = p.numel()
        total_params += n
        if p.requires_grad:
            trainable_params += n
        try:
            local_sq_norm += float(p.detach().float().norm(2).item() ** 2)
        except Exception:
            pass
    return total_params, trainable_params, local_sq_norm


@contextmanager
def skip_random_init():
    """
    Context manager to skip random weight initialization when loading pretrained models.
    """
    try:
        mod = __import__("transformers.initialization", fromlist=["_init_weights"])
    except ImportError:
        mod = __import__("transformers.modeling_utils", fromlist=["_init_weights"])
    prev = getattr(mod, "_init_weights", True)
    mod._init_weights = False
    try:
        yield
    finally:
        mod._init_weights = prev


def resolve_trust_remote_code(pretrained_model_name_or_path):
    """
    Whitelist NVIDIA models to allow remote code execution.

    Args:
        pretrained_model_name_or_path (str): The name or path of the pretrained model.

    Returns:
        bool: True if the model should be loaded with trust_remote_code, False otherwise.
    """
    if not pretrained_model_name_or_path:
        return False
    # pretrained_model_name_or_path can be something like nvidia/NVIDIA-Nemotron-Nano-9B-v2
    return not os.path.isdir(pretrained_model_name_or_path) and pretrained_model_name_or_path.startswith("nvidia/")


def print_trainable_parameters(model: nn.Module) -> tuple[int, int]:
    """Print the number of trainable parameters in the model.

    Args:
        model: Model to analyze

    Returns:
        trainable_params: int
        total_params: int
    """
    total_params, trainable_params, local_sq_norm = _get_model_param_stats(model)

    try:
        # TODO(@akoumparouli): make this sharding aware.
        local_sq_norm = float(local_sq_norm**0.5)
        trainable_pct = (100.0 * trainable_params / total_params) if total_params > 0 else 0.0

        logging.info("Model summary:")
        logging.info("--------------------------------")
        logging.info(f"Trainable parameters: {trainable_params:,}")
        logging.info(f"Total parameters: {total_params:,}")
        logging.info(f"Trainable parameters percentage: {trainable_pct:.2f}%")
        logging.info(f"Param L2 norm: {local_sq_norm:.4f}")
        logging.info("--------------------------------")
    except Exception:
        logging.info("Model summary: <unavailable>")

    return trainable_params, total_params


def _freeze_module_by_attribute_and_patterns(model, attribute_name, name_patterns):
    """Helper function to freeze parameters by attribute name and name patterns.

    Args:
        model: The model to apply freezing to.
        attribute_name: Name of the model attribute to freeze (e.g., 'vision_tower').
        name_patterns: List of patterns to match in module names.
    """
    # Freeze by attribute name
    if hasattr(model, attribute_name):
        for param in getattr(model, attribute_name).parameters():
            param.requires_grad = False

    # Freeze by name patterns
    for name, module in model.named_modules():
        if any(pattern in name.lower() for pattern in name_patterns):
            for param in module.parameters():
                param.requires_grad = False


def apply_parameter_freezing(model, freeze_config):
    """Apply parameter freezing based on configuration.

    Args:
        model: The model to apply freezing to.
        freeze_config: Configuration dict specifying what to freeze.

    freeze_config can contain:
        - freeze_vision_tower: bool (default True)
        - freeze_audio_tower: bool (default False)
        - freeze_language_model: bool (default False)
    """
    freeze_vision_tower = freeze_config.get("freeze_vision_tower", True)
    freeze_audio_tower = freeze_config.get("freeze_audio_tower", False)
    freeze_language_model = freeze_config.get("freeze_language_model", False)

    # Freeze vision tower
    if freeze_vision_tower:
        _freeze_module_by_attribute_and_patterns(model, "vision_tower", ["vision", "visual", "image_encoder"])

    # Freeze audio tower
    if freeze_audio_tower:
        _freeze_module_by_attribute_and_patterns(model, "audio_tower", ["audio", "audio_encoder"])

    # Freeze language model backbone
    if freeze_language_model:
        _freeze_module_by_attribute_and_patterns(model, "language_model", ["language", "text", "llm"])


def squeeze_input_for_thd(input_ids, position_ids, padding_mask, attn_kwargs, seqlens_padding_value=-1000):
    """
    Squeeze batch dimension and prepare inputs for THD (total, hidden, depth) format.

    This function removes the batch dimension from input tensors and processes attention
    kwargs for use with Transformer Engine's THD format. It's typically used when the
    batch has already been converted to THD format (with batch_size=1 as a placeholder
    dimension) and that dimension needs to be removed.

    The function performs three key operations:
    1. Removes the batch dimension (dim 0) from input tensors
    2. Filters out padding values from cumulative sequence length tensors
    3. Converts max_seqlen from tensor to scalar if needed

    Args:
        input_ids (torch.Tensor): Input token IDs with shape [1, total_tokens] or
            [1, total_tokens, hidden_dim]. The first dimension will be squeezed.
        position_ids (torch.Tensor): Position IDs with shape [1, total_tokens].
            The first dimension will be squeezed.
        padding_mask (torch.Tensor): Padding mask with shape [1, total_tokens].
            The first dimension will be squeezed.
        attn_kwargs (dict): Dictionary of attention-related tensors. May contain:
            - cu_seqlens: Cumulative sequence lengths [1, num_seqs+1]
            - cu_seqlens_padded: Cumulative padded sequence lengths [1, num_seqs+1]
            - max_seqlen: Maximum sequence length (tensor or int)
            - Other attention parameters (will be squeezed if tensors)
        seqlens_padding_value (int): Sentinel value used to indicate padding in
            cu_seqlens and cu_seqlens_padded tensors. These values will be filtered
            out. Default: -1000.

    Returns:
        tuple: A tuple containing:
            - input_ids (torch.Tensor): Input IDs with batch dimension removed [total_tokens]
                or [total_tokens, hidden_dim]
            - position_ids (torch.Tensor): Position IDs with batch dimension removed [total_tokens]
            - padding_mask (torch.Tensor): Padding mask with batch dimension removed [total_tokens]
            - attn_kwargs (dict): Updated attention kwargs with:
                - Batch dimensions removed from all tensor values
                - Padding values filtered from cu_seqlens and cu_seqlens_padded
                - max_seqlen converted to scalar if it was a tensor

    Example:
        >>> input_ids = torch.tensor([[1, 2, 3, 4, 5]])  # [1, 5]
        >>> position_ids = torch.tensor([[0, 1, 2, 3, 4]])  # [1, 5]
        >>> padding_mask = torch.tensor([[False, False, False, False, False]])  # [1, 5]
        >>> attn_kwargs = {
        ...     'cu_seqlens': torch.tensor([[0, 3, 5, -1000]]),  # [1, 4] with padding
        ...     'cu_seqlens_padded': torch.tensor([[0, 3, 5, -1000]]),
        ...     'max_seqlen': torch.tensor([3])
        ... }
        >>> ids, pos, mask, kwargs = squeeze_input_for_thd(
        ...     input_ids, position_ids, padding_mask, attn_kwargs
        ... )
        >>> ids.shape
        torch.Size([5])
        >>> kwargs['cu_seqlens']  # Padding value filtered out
        tensor([0, 3, 5])
        >>> kwargs['max_seqlen']  # Converted to scalar
        3

    Note:
        This function modifies attn_kwargs in-place. If you need to preserve the original
        dictionary, pass a copy.
    """
    input_ids = input_ids.squeeze(0)
    position_ids = position_ids.squeeze(0)
    if padding_mask is not None:
        padding_mask = padding_mask.squeeze(0)
    for key, value in attn_kwargs.items():
        if isinstance(value, torch.Tensor):
            attn_kwargs[key] = value.squeeze(0)
        if key in ["cu_seqlens", "cu_seqlens_padded"]:
            attn_kwargs[key] = value[value != seqlens_padding_value].contiguous()
        if key == "max_seqlen" and isinstance(value, torch.Tensor):
            attn_kwargs[key] = value.item()

    return input_ids, position_ids, padding_mask, attn_kwargs


# taken and edited from https://github.com/huggingface/transformers/blob/32a58e31463e238c967207bf73772490c353551a/src/transformers/integrations/accelerate.py#L53-L158
@contextmanager
def init_empty_weights():
    """
    A context manager under which models are initialized with all parameters on the specified device.

    Args:
        device (`torch.device`):
            Device to initialize all parameters on.

    Example:

    ```python
    import torch.nn as nn
    from nemo_automodel.components.utils.model_utils import init_empty_weights

    with init_empty_weights():
        tst = nn.Linear(100, 100)  # on `cuda` device
    ```
    """
    device = torch.device("meta")
    fp8_parameter_mapping = {
        "_linear_mm_config": "linear_mm_config",
        "_dtype": "dtype",
        "_precomputed_scale": "precomputed_scale",
    }
    old_register_parameter = nn.Module.register_parameter

    def register_empty_parameter(module, name, param):
        old_register_parameter(module, name, param)
        if param is not None:
            param_cls = type(module._parameters[name])
            if HAVE_TORCHAO and isinstance(
                module._parameters[name], torch_ao.float8.fsdp_utils.WeightWithDynamicFloat8CastTensor
            ):
                kwargs = {}
                for k in module._parameters[name].__dict__:
                    if k in fp8_parameter_mapping:
                        kwargs[fp8_parameter_mapping[k]] = getattr(module._parameters[name], k)
                is_hf_initialized = kwargs.pop("_is_hf_initialized", None)
            else:
                # Standard nn.Parameter only accepts requires_grad, not arbitrary __dict__ attributes
                # (e.g., TransformerEngine sets tensor_model_parallel on weights)
                if param_cls is nn.Parameter:
                    kwargs = {"requires_grad": param.requires_grad}
                    is_hf_initialized = None
                else:
                    kwargs = module._parameters[name].__dict__.copy()
                    kwargs["requires_grad"] = param.requires_grad
                    is_hf_initialized = kwargs.pop("_is_hf_initialized", None)
            module._parameters[name] = param_cls(module._parameters[name].to(device), **kwargs)
            if is_hf_initialized is not None:
                setattr(module._parameters[name], "_is_hf_initialized", is_hf_initialized)

    try:
        nn.Module.register_parameter = register_empty_parameter
        yield
    finally:
        nn.Module.register_parameter = old_register_parameter



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""NeMo Auto Model classes.

Drop-in replacements for ``transformers.AutoModelFor*`` that add custom-kernel
patching, distributed infrastructure, PEFT, quantization, and checkpointing.

Heavy-lifting helpers live in sibling modules:

* ``kernel_patches`` -- SDPA / Liger kernel patching
* ``model_init`` -- model class resolution and instantiation
* ``infrastructure`` -- MeshContext, sharding, PEFT/quant application
"""

import gc
import logging
from contextlib import nullcontext
from typing import TYPE_CHECKING, List, Optional, Union

import torch
from torch.nn.attention import SDPBackend

from nemo_automodel.shared.torch_patches import apply_torch_patches

apply_torch_patches()
from huggingface_hub import constants as hf_constants  # noqa: E402
from transformers import (  # noqa: E402
    AutoModelForCausalLM,
    AutoModelForImageTextToText,
    AutoModelForMultimodalLM,
    AutoModelForSequenceClassification,
    AutoModelForTextToWaveform,
    PreTrainedModel,
)
from transformers.initialization import no_init_weights  # noqa: E402
from transformers.models.auto.auto_factory import _BaseAutoModelClass  # noqa: E402
from transformers.utils import ContextManagers  # noqa: E402

from nemo_automodel.components.distributed.config import (  # noqa: E402
    DistributedConfig,
)
from nemo_automodel.components.distributed.ddp import DDPManager  # noqa: E402
from nemo_automodel.components.distributed.init_utils import get_world_size_safe  # noqa: E402
from nemo_automodel.components.distributed.megatron_fsdp import MegatronFSDPManager  # noqa: E402
from nemo_automodel.components.distributed.pipelining.config import PipelineConfig  # noqa: E402
from nemo_automodel.components.moe.config import MoEParallelizerConfig  # noqa: E402
from nemo_automodel.components.quantization.qat import QATConfig  # noqa: E402
from nemo_automodel.components.utils.model_utils import (  # noqa: E402
    init_empty_weights,
    resolve_trust_remote_code,
)
from nemo_automodel.shared.utils import dtype_from_str  # noqa: E402

if TYPE_CHECKING:
    from torch.distributed.device_mesh import DeviceMesh

    from nemo_automodel.components.quantization.fp8 import FP8Config
    from nemo_automodel.components.utils.compile_utils import CompileConfig

#  Re-exports from sibling modules (backward compatibility)
# Backward-compat shim for trust_remote_code models (e.g. DeciLM)
# that import NEED_SETUP_CACHE_CLASSES_MAPPING from transformers.generation.utils.
import transformers.generation.utils as _gen_utils  # noqa: E402

from nemo_automodel._transformers.infrastructure import (
    MeshContext,
    apply_model_infrastructure,
    instantiate_infrastructure,
)
from nemo_automodel._transformers.kernel_patches import (
    DEFAULT_ATTN_IMPLEMENTATION,
    _apply_preload_overrides,
    _get_next_fallback_attn,
    _patch_attention,
    _patch_liger_kernel,
    _verify_sdpa_support,
)
from nemo_automodel._transformers.model_init import (
    _consume_config_overrides,
    _init_model,
    get_hf_config,
    get_is_hf_model,
    no_hf_meta_device,
)

if not hasattr(_gen_utils, "NEED_SETUP_CACHE_CLASSES_MAPPING"):
    from transformers.cache_utils import StaticCache

    _gen_utils.NEED_SETUP_CACHE_CLASSES_MAPPING = {"static": StaticCache}


logger = logging.getLogger(__name__)


class _BaseNeMoAutoModelClass(_BaseAutoModelClass):
    """
    Drop-in replacement for ``_BaseAutoModelClass`` that includes custom-kernels.

    The class only overrides ``from_pretrained`` and ``from_config`` to add the
    optional ``use_liger_kernel`` flag.  If the flag is ``True`` (default) and
    the Liger kernel is available, the model's attention layers are
    monkey-patched in place.  If patching fails for any reason, the call is
    retried once with ``use_liger_kernel=False`` so that users still obtain a
    functional model.


    TODO(@akoumpa): extend this beyond liger_kernel.

    Notes:
    -----
    - No changes are made to the model's public API; forward signatures,
      generation utilities, and weight shapes remain identical.
    - Only decoder-style (causal) architectures are currently supported by the
      Liger patch.  Unsupported models will silently fall back.
    """

    @classmethod
    def _from_pretrained_parent_class(cls, *args, **kwargs):
        name = cls.__name__
        if name.startswith("NeMo"):
            cls.__name__ = name[4:]
        model = super().from_pretrained(*args, **kwargs)
        cls.__name__ = name
        return model

    @classmethod
    def _from_config_parent_class(cls, *args, **kwargs):
        name = cls.__name__
        if name.startswith("NeMo"):
            cls.__name__ = name[4:]
        model = super().from_config(*args, **kwargs)
        cls.__name__ = name
        return model

    @classmethod
    def _build_model(
        cls,
        pretrained_model_name_or_path_or_config,
        *model_args,
        is_hf_model,
        use_liger_kernel,
        use_sdpa_patching,
        sdpa_method,
        torch_dtype,
        attn_implementation,
        quantization_config,
        force_hf,
        model_wrapper,
        autopipeline,
        parallelize_fn,
        qat_quantizer,
        mesh,
        loss_fn,
        peft_config,
        fp8_config,
        compile_config,
        load_base_model,
        **kwargs,
    ):
        """Shared model building logic for ``from_pretrained`` and ``from_config``.

        Handles pre-load overrides, meta-device initialization, model init with
        attention-fallback retry, kernel patching (Liger, SDPA) with retry, and
        full infrastructure application (sharding, PEFT, quantization, checkpointing).

        All caller-specific setup (config resolution, infrastructure instantiation,
        ``is_hf_model`` determination) is done by ``from_pretrained`` / ``from_config``
        before delegating here.
        """
        # Extract values consumed by pop; preserve them for retry.
        kwargs = dict(kwargs)  # Defensive copy so retries get clean state
        has_packed_sequence = kwargs.pop("has_packed_sequence", False)
        freeze_config = kwargs.pop("freeze_config", None)
        cache_dir = kwargs.pop("cache_dir", hf_constants.HF_HUB_CACHE)

        def _retry(**override):
            """Re-enter ``_build_model`` with overridden parameters."""
            retry_kwargs = {
                **kwargs,
                "has_packed_sequence": has_packed_sequence,
                "freeze_config": freeze_config,
                "cache_dir": cache_dir,
            }
            return cls._build_model(
                pretrained_model_name_or_path_or_config,
                *model_args,
                is_hf_model=is_hf_model,
                use_liger_kernel=override.get("use_liger_kernel", use_liger_kernel),
                use_sdpa_patching=override.get("use_sdpa_patching", use_sdpa_patching),
                sdpa_method=sdpa_method,
                torch_dtype=torch_dtype,
                attn_implementation=override.get("attn_implementation", attn_implementation),
                quantization_config=quantization_config,
                force_hf=force_hf,
                model_wrapper=model_wrapper,
                autopipeline=autopipeline,
                parallelize_fn=parallelize_fn,
                qat_quantizer=qat_quantizer,
                mesh=mesh,
                loss_fn=loss_fn,
                peft_config=peft_config,
                fp8_config=fp8_config,
                compile_config=compile_config,
                load_base_model=load_base_model,
                **retry_kwargs,
            )

        if is_hf_model:
            attn_implementation, use_liger_kernel = _apply_preload_overrides(
                mesh.tp_size,
                mesh.cp_size,
                has_packed_sequence,
                attn_implementation,
                use_liger_kernel,
            )
        device = torch.cuda.current_device()

        # Use meta device initialization when:
        # - Not using MegatronFSDPManager or DDPManager (they handle their own initialization)
        # - AND either multi-GPU (world_size > 1) or single-GPU custom model (not HF)
        # - AND not using quantization (we let HF handle BitsAndBytes; don't init meta device)
        is_meta_device = all(
            [
                not isinstance(model_wrapper, (MegatronFSDPManager, DDPManager)),
                get_world_size_safe() > 1 or not is_hf_model,
                quantization_config is None,
            ]
        )
        init_ctx = ContextManagers([no_init_weights(), init_empty_weights()]) if is_meta_device else nullcontext()

        model = None  # Ensure 'model' is always bound for the except handler
        is_custom_model = None
        try:
            with init_ctx:
                is_custom_model, model = _init_model(
                    cls,
                    pretrained_model_name_or_path_or_config,
                    attn_implementation,
                    torch_dtype,
                    quantization_config,
                    force_hf,
                    *model_args,
                    **kwargs,
                )
        except NotImplementedError as e:
            if "Cannot copy out of meta tensor" in str(e) and is_meta_device:
                logger.warning(
                    "Model init hit 'Cannot copy out of meta tensor' (e.g. buffer created with meta but "
                    "called .to(device)); retrying without meta device.",
                )
                del model
                model = None
                gc.collect()
                is_meta_device = False
                with ContextManagers([no_init_weights(), no_hf_meta_device()]):
                    is_custom_model, model = _init_model(
                        cls,
                        pretrained_model_name_or_path_or_config,
                        attn_implementation,
                        torch_dtype,
                        quantization_config,
                        force_hf,
                        *model_args,
                        **kwargs,
                    )
            else:
                raise
        except ValueError as e:
            if "does not support" in str(e):
                del model
                attn_implementation = _get_next_fallback_attn(attn_implementation)
                logger.warning("Falling back to %s attention.", attn_implementation)
                return _retry(attn_implementation=attn_implementation)
            raise

        # Kernel patching
        try:
            if use_liger_kernel and not is_custom_model:
                model = _patch_liger_kernel(model)
        except RuntimeError:
            logger.warning("Retrying without Liger kernels.")
            del model
            gc.collect()
            return _retry(use_liger_kernel=False)

        try:
            if use_sdpa_patching and not is_custom_model:
                model = _patch_attention(model, sdpa_method)  # noqa: F821
        except Exception:
            logger.warning("Retrying without SDPA patching.")
            return _retry(use_sdpa_patching=False)

        # Resolve pretrained path for checkpoint loading
        is_pretrained = isinstance(pretrained_model_name_or_path_or_config, str)
        pretrained_path = (
            pretrained_model_name_or_path_or_config
            if is_pretrained
            else getattr(pretrained_model_name_or_path_or_config, "name_or_path", "")
        )

        if is_hf_model:
            _verify_sdpa_support(model, mesh.cp_size)

        model = apply_model_infrastructure(
            model=model,
            pretrained_model_name_or_path=pretrained_path,
            mesh=mesh,
            peft_config=peft_config,
            quantization_config=quantization_config,
            fp8_config=fp8_config,
            qat_quantizer=qat_quantizer,
            loss_fn=loss_fn,
            autopipeline=autopipeline,
            parallelize_fn=parallelize_fn,
            model_wrapper=model_wrapper,
            is_meta_device=is_meta_device,
            device=device,
            compile_config=compile_config,
            load_base_model=load_base_model,
            cache_dir=cache_dir,
            freeze_config=freeze_config,
        )

        return model

    @classmethod
    def from_pretrained(
        cls,
        pretrained_model_name_or_path,
        *model_args,
        use_liger_kernel: bool = True,
        use_sdpa_patching: bool = True,
        sdpa_method: Optional[List[SDPBackend]] = None,
        torch_dtype="auto",
        attn_implementation: str = DEFAULT_ATTN_IMPLEMENTATION,
        quantization_config=None,
        force_hf: bool = False,
        device_mesh: Optional["DeviceMesh"] = None,
        moe_mesh: Optional["DeviceMesh"] = None,
        tp_plan: Optional[dict] = None,
        distributed_config: Optional[DistributedConfig] = None,
        pipeline_config: Optional[PipelineConfig] = None,
        qat_config: Optional[QATConfig] = None,
        moe_config: Optional[MoEParallelizerConfig] = None,
        activation_checkpointing: bool = False,
        peft_config: Optional[dict] = None,
        fp8_config: Optional["FP8Config"] = None,
        compile_config: Optional["CompileConfig"] = None,
        **kwargs,
    ) -> PreTrainedModel:
        """
        Instantiate and (optionally) patch a causal-language model.

        This is a light wrapper around
        `transformers.AutoModelForCausalLM.from_pretrained` that can
        automatically apply Liger and/or SDPA (scaled-dot-product
        attention) kernel optimizations, as well as PEFT, quantization,
        and distributed parallelism.

        Args:
            pretrained_model_name_or_path (str | os.PathLike): Hugging Face
                hub repo ID or local path accepted by
                `AutoModelForCausalLM.from_pretrained`.
            *model_args: Positional arguments forwarded verbatim to
                `AutoModelForCausalLM.from_pretrained`.
            use_liger_kernel (bool, default=True): If `True`, try to patch
                the model with Liger kernels for faster inference/training.
            use_sdpa_patching (bool, default=True): If `True`, patch the
                model with SDPA-based attention optimizations.
            sdpa_method (list[SDPBackend] | None, optional): Explicit list of
                SDPA back-ends to consider when `use_sdpa_patching=True`.
            torch_dtype (str | torch.dtype | Literal["auto"], default="auto"):
                Data type passed to the underlying `from_pretrained` call.
            attn_implementation (str, optional):
                Specifies which attention implementation to use (e.g.,
                ``"flash_attention_2"``, ``"eager"``). Only applied when the
                base model supports this kwarg. Defaults to ``"flash_attention_2"``,
                if flash attention is not available, defaults to ``"sdpa"``.
            quantization_config (optional): BitsAndBytesConfig configuration object that
                specifies all quantization settings. If provided, quantization
                will be applied to the model.
            force_hf (bool, default=False): If `True`, force the use of HF model implementation.
                If `False`, the model will be loaded using the custom model implementation if available.
            device_mesh (DeviceMesh | None, optional): Pre-created device mesh for
                distributed training. Parallelism sizes (tp, pp, cp, ep) are inferred
                from this. Default: None.
            moe_mesh (DeviceMesh | None, optional): FSDP2-only. Device mesh for expert
                parallelism. ep_size is inferred from this. Default: None.
            tp_plan (dict | None, optional): Custom tensor parallel plan. If provided,
                overrides the tp_plan on distributed_config. Default: None.
            distributed_config (FSDP2Config | MegatronFSDPConfig | DDPConfig | None, optional):
                Strategy-specific distributed training configuration. Default: None.
            pipeline_config (PipelineConfig | None, optional): Pipeline parallelism
                configuration including loss_fn. Default: None.
            qat_config (QATConfig | None, optional): Quantization-Aware Training
                configuration. Default: None.
            moe_config (MoEParallelizerConfig | None, optional): MoE parallelizer
                configuration. Default: None.
            activation_checkpointing (bool, default=False): Enable activation checkpointing
                for transformer blocks to reduce memory usage. Default: False.
            peft_config (dict | None, optional): PEFT/LoRA configuration dictionary.
                If provided, LoRA adapters will be applied to the model. Default: None.
            fp8_config (FP8Config | None, optional): FP8 quantization configuration.
                If provided, FP8 quantization will be applied. Default: None.
            compile_config (CompileConfig | None, optional): Configuration for torch.compile.
                If provided, the model will be compiled. Default: None.
            **kwargs: Additional keyword arguments. Notable ones include:
                - has_packed_sequence (bool): Whether using packed sequences. Default: False.
                - cache_dir (str): Cache directory for model weights.

        Returns:
            transformers.PreTrainedModel: The loaded (and possibly patched)
            model instance with all infrastructure applied.
        """
        if tp_plan is not None and distributed_config is not None:
            distributed_config.tp_plan = tp_plan

        mesh = MeshContext.from_meshes(device_mesh, moe_mesh)

        model_wrapper, autopipeline, parallelize_fn, qat_quantizer = instantiate_infrastructure(
            distributed_config=distributed_config,
            pipeline_config=pipeline_config,
            qat_config=qat_config,
            moe_config=moe_config,
            activation_checkpointing=activation_checkpointing,
            device=torch.device("cuda", torch.cuda.current_device()),
            mesh=mesh,
        )
        loss_fn = pipeline_config.loss_fn if pipeline_config is not None else None

        try:
            hf_config = get_hf_config(pretrained_model_name_or_path, attn_implementation, **kwargs)
        except Exception as e:
            if "does not support" in str(e):
                attn_implementation = _get_next_fallback_attn(attn_implementation)
                logger.warning("Config rejected attention implementation, falling back to %s.", attn_implementation)
                hf_config = get_hf_config(pretrained_model_name_or_path, attn_implementation, **kwargs)
            else:
                raise
        is_hf_model = get_is_hf_model(hf_config, force_hf)

        return cls._build_model(
            pretrained_model_name_or_path,
            *model_args,
            is_hf_model=is_hf_model,
            use_liger_kernel=use_liger_kernel,
            use_sdpa_patching=use_sdpa_patching,
            sdpa_method=sdpa_method,
            torch_dtype=torch_dtype,
            attn_implementation=attn_implementation,
            quantization_config=quantization_config,
            force_hf=force_hf,
            model_wrapper=model_wrapper,
            autopipeline=autopipeline,
            parallelize_fn=parallelize_fn,
            qat_quantizer=qat_quantizer,
            mesh=mesh,
            loss_fn=loss_fn,
            peft_config=peft_config,
            fp8_config=fp8_config,
            compile_config=compile_config,
            load_base_model=True,
            **kwargs,
        )

    @classmethod
    def from_config(
        cls,
        config,
        *model_args,
        use_liger_kernel: bool = True,
        use_sdpa_patching: bool = True,
        sdpa_method: Optional[List[SDPBackend]] = None,
        torch_dtype: Union[str, torch.dtype] = "auto",
        attn_implementation: str = DEFAULT_ATTN_IMPLEMENTATION,
        quantization_config=None,
        force_hf: bool = False,
        device_mesh: Optional["DeviceMesh"] = None,
        moe_mesh: Optional["DeviceMesh"] = None,
        tp_plan: Optional[dict] = None,
        distributed_config: Optional[DistributedConfig] = None,
        pipeline_config: Optional[PipelineConfig] = None,
        qat_config: Optional[QATConfig] = None,
        moe_config: Optional[MoEParallelizerConfig] = None,
        activation_checkpointing: bool = False,
        peft_config: Optional[dict] = None,
        fp8_config: Optional["FP8Config"] = None,
        compile_config: Optional["CompileConfig"] = None,
        **kwargs,
    ) -> PreTrainedModel:
        """
        Instantiate a model from a ``transformers.PretrainedConfig`` (no pretrained
        weights). Accepts the same infrastructure arguments as ``from_pretrained``.

        See ``from_pretrained`` for full parameter documentation.

        Args:
            config (transformers.PretrainedConfig | str):
                The configuration object used to build the model.
                If config is passed as a string (e.g., model-id / local checkpoint),
                it will create a config internally using AutoConfig.
            torch_dtype (str | torch.dtype, default="auto"):
                Data type for model parameters. If "auto", defaults to ``torch.bfloat16``.
        """
        if tp_plan is not None and distributed_config is not None:
            distributed_config.tp_plan = tp_plan

        mesh = MeshContext.from_meshes(device_mesh, moe_mesh)

        # Only instantiate infrastructure when distributed_config is provided
        model_wrapper = autopipeline = parallelize_fn = qat_quantizer = None
        loss_fn = None
        if distributed_config is not None:
            model_wrapper, autopipeline, parallelize_fn, qat_quantizer = instantiate_infrastructure(
                distributed_config=distributed_config,
                pipeline_config=pipeline_config,
                qat_config=qat_config,
                moe_config=moe_config,
                activation_checkpointing=activation_checkpointing,
                device=torch.device("cuda", torch.cuda.current_device()),
                mesh=mesh,
            )
            if pipeline_config is not None:
                loss_fn = pipeline_config.loss_fn

        torch_dtype = dtype_from_str(torch_dtype) if torch_dtype != "auto" else torch.bfloat16
        name_or_path = config if isinstance(config, str) else getattr(config, "name_or_path", None)
        kwargs["trust_remote_code"] = kwargs.get(
            "trust_remote_code", resolve_trust_remote_code(name_or_path) if name_or_path else False
        )
        if isinstance(config, str):
            try:
                config = get_hf_config(config, attn_implementation, **kwargs)
            except Exception as e:
                if "does not support" in str(e):
                    attn_implementation = _get_next_fallback_attn(attn_implementation)
                    logger.warning("Config rejected attention implementation, falling back to %s.", attn_implementation)
                    config = get_hf_config(config, attn_implementation, **kwargs)
                else:
                    raise
        _consume_config_overrides(config, kwargs)
        is_hf_model = get_is_hf_model(config, force_hf)

        return cls._build_model(
            config,
            *model_args,
            is_hf_model=is_hf_model,
            use_liger_kernel=use_liger_kernel,
            use_sdpa_patching=use_sdpa_patching,
            sdpa_method=sdpa_method,
            torch_dtype=torch_dtype,
            attn_implementation=attn_implementation,
            quantization_config=quantization_config,
            force_hf=force_hf,
            model_wrapper=model_wrapper,
            autopipeline=autopipeline,
            parallelize_fn=parallelize_fn,
            qat_quantizer=qat_quantizer,
            mesh=mesh,
            loss_fn=loss_fn,
            peft_config=peft_config,
            fp8_config=fp8_config,
            compile_config=compile_config,
            load_base_model=kwargs.pop("load_base_model", False),
            **kwargs,
        )


#  Concrete Auto-Model classes
class NeMoAutoModelForCausalLM(_BaseNeMoAutoModelClass, AutoModelForCausalLM):
    """
    Drop-in replacement for ``transformers.AutoModelForCausalLM`` that includes custom-kernels.

    The class only overrides ``from_pretrained`` and ``from_config`` to add the
    optional ``use_liger_kernel`` flag.  If the flag is ``True`` (default) and
    the Liger kernel is available, the model's attention layers are
    monkey-patched in place.  If patching fails for any reason, the call is
    retried once with ``use_liger_kernel=False`` so that users still obtain a
    functional model.


    TODO(@akoumpa): extend this beyond liger_kernel.

    Notes:
    -----
    - No changes are made to the model's public API; forward signatures,
      generation utilities, and weight shapes remain identical.
    - Only decoder-style (causal) architectures are currently supported by the
      Liger patch.  Unsupported models will silently fall back.

    Examples:
    --------
    >>> model = NeMoAutoModelForCausalLM.from_pretrained("gpt2")            # try Liger
    >>> model = NeMoAutoModelForCausalLM.from_pretrained(
    ...     "gpt2", use_liger_kernel=False)                                 # skip Liger
    """

    pass


class NeMoAutoModelForImageTextToText(_BaseNeMoAutoModelClass, AutoModelForImageTextToText):
    """Drop-in replacement for ``transformers.AutoModelForImageTextToText`` with custom-kernels.

    The class only overrides ``from_pretrained`` and ``from_config`` to add the
    optional ``use_liger_kernel`` flag.  If the flag is ``True`` (default) and
    the Liger kernel is available, the model's attention layers are
    monkey-patched in place.  If patching fails for any reason, the call is
    retried once with ``use_liger_kernel=False`` so that users still obtain a
    functional model.


    @akoumpa: currently only supporting liger_kernel for demonstration purposes.

    Notes:
    -----
    - No changes are made to the model's public API; forward signatures,
      generation utilities, and weight shapes remain identical.
    - Only decoder-style (causal) architectures are currently supported by the
      Liger patch.  Unsupported models will silently fall back.

    Examples:
    --------
    >>> model = NeMoAutoModelForImageTextToText.from_pretrained("Qwen/Qwen2.5-VL-3B-Instruct") # try Liger
    >>> model = NeMoAutoModelForImageTextToText.from_pretrained(
    ...     "Qwen/Qwen2.5-VL-3B-Instruct", use_liger_kernel=False)                            # skip Liger
    """

    pass


class NeMoAutoModelForMultimodalLM(_BaseNeMoAutoModelClass, AutoModelForMultimodalLM):
    """Drop-in replacement for ``transformers.AutoModelForMultimodalLM`` with custom-kernels."""

    pass


class NeMoAutoModelForSequenceClassification(_BaseNeMoAutoModelClass, AutoModelForSequenceClassification):
    """Drop-in replacement for ``transformers.AutoModelForSequenceClassification`` with custom-kernels.

    The class only overrides ``from_pretrained`` and ``from_config`` to add the
    optional ``use_liger_kernel`` flag.  If the flag is ``True`` (default) and
    the Liger kernel is available, the model's attention layers are
    monkey-patched in place.  If patching fails for any reason, the call is
    retried once with ``use_liger_kernel=False`` so that users still obtain a
    functional model.


    @akoumpa: currently only supporting liger_kernel for demonstration purposes.

    Notes:
    -----
    - No changes are made to the model's public API; forward signatures,
      generation utilities, and weight shapes remain identical.
    - Only decoder-style (causal) architectures are currently supported by the
      Liger patch.  Unsupported models will silently fall back.

    Examples:
    --------
    >>> model = NeMoAutoModelForSequenceClassification.from_pretrained("bert-base-uncased") # try Liger
    >>> model = NeMoAutoModelForSequenceClassification.from_pretrained(
    ...     "bert-base-uncased", use_liger_kernel=False)                            # skip Liger
    """

    pass


class NeMoAutoModelForTextToWaveform(_BaseNeMoAutoModelClass, AutoModelForTextToWaveform):
    """Drop-in replacement for ``transformers.AutoModelForTextToWaveform`` with custom-kernels.

    The class only overrides ``from_pretrained`` and ``from_config`` to add the
    optional ``use_liger_kernel`` flag.  If the flag is ``True`` (default) and
    the Liger kernel is available, the model's attention layers are
    monkey-patched in place.  If patching fails for any reason, the call is
    retried once with ``use_liger_kernel=False`` so that users still obtain a
    functional model.


    @akoumpa: currently only supporting liger_kernel for demonstration purposes.

    Notes:
    -----
    - No changes are made to the model's public API; forward signatures,
      generation utilities, and weight shapes remain identical.
    - Only decoder-style (causal) architectures are currently supported by the
      Liger patch.  Unsupported models will silently fall back.

    Examples:
    --------
    >>> model = NeMoAutoModelForTextToWaveform.from_pretrained("facebook/musicgen-small") # try Liger
    >>> model = NeMoAutoModelForTextToWaveform.from_pretrained(
    ...     "facebook/musicgen-small", use_liger_kernel=False)                            # skip Liger
    """

    pass



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Llama Bidirectional Model for NeMo AutoModel.

This module provides a bidirectional attention variant of Llama that is useful
for embedding and retrieval tasks. Unlike the standard causal Llama model,
this version can attend to all tokens bidirectionally.
"""

import copy
import os
from dataclasses import dataclass
from typing import Dict, List, Optional, Tuple, Union

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor
from torch.nn import BCEWithLogitsLoss, CrossEntropyLoss, MSELoss
from transformers import PreTrainedModel
from transformers.cache_utils import Cache, DynamicCache
from transformers.modeling_outputs import (
    BaseModelOutputWithPast,
    ModelOutput,
    SequenceClassifierOutputWithPast,
)
from transformers.models.llama.configuration_llama import LlamaConfig
from transformers.models.llama.modeling_llama import (
    LlamaForSequenceClassification,
    LlamaModel,
)
from transformers.processing_utils import Unpack
from transformers.utils import TransformersKwargs, auto_docstring, logging

try:
    from nemo_automodel.components.models.biencoder.state_dict_adapter import BiencoderStateDictAdapter
except ImportError:
    BiencoderStateDictAdapter = object

try:
    from nemo_automodel.shared.import_utils import get_check_model_inputs_decorator

    check_model_inputs = get_check_model_inputs_decorator()
except ImportError:
    # Fallback to no-op decorator if import fails
    def check_model_inputs(func):
        return func


logger = logging.get_logger(__name__)


def contrastive_scores_and_labels(
    query: torch.Tensor, key: torch.Tensor, current_train_n_passages: int
) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Compute contrastive scores and labels without in-batch negatives.

    Args:
        query: Query embeddings [batch_size, hidden_dim]
        key: Key/passage embeddings [batch_size * n_passages, hidden_dim]
        current_train_n_passages: Number of passages per query

    Returns:
        Tuple of (scores, labels) where scores is [batch_size, n_passages]
        and labels is [batch_size] of zeros (positive is first passage)
    """
    assert key.shape[0] % query.shape[0] == 0, "{} % {} > 0".format(key.shape[0], query.shape[0])
    query_shape = query.shape
    repeated_query = query.repeat(1, 1, current_train_n_passages).reshape(
        query_shape[0] * current_train_n_passages, query_shape[1]
    )
    qk = torch.sum(repeated_query * key, dim=-1).reshape(query_shape[0], current_train_n_passages)
    labels = torch.zeros(query_shape[0], dtype=torch.long, device=query.device)
    return qk, labels


def pool(last_hidden_states: torch.Tensor, attention_mask: torch.Tensor, pool_type: str) -> torch.Tensor:
    """
    Pool hidden states using the specified pooling method.

    Args:
        last_hidden_states: Hidden states from the model [batch_size, seq_len, hidden_size]
        attention_mask: Attention mask [batch_size, seq_len]
        pool_type: Type of pooling to apply

    Returns:
        Pooled embeddings [batch_size, hidden_size]
    """
    last_hidden = last_hidden_states.masked_fill(~attention_mask[..., None].bool(), 0.0)

    if pool_type == "avg":
        emb = last_hidden.sum(dim=1) / attention_mask.sum(dim=1)[..., None]
    elif pool_type == "weighted_avg":
        emb = last_hidden.sum(dim=1)
    elif pool_type == "cls":
        emb = last_hidden[:, 0]
    elif pool_type == "last":
        left_padding = attention_mask[:, -1].sum() == attention_mask.shape[0]
        if left_padding:
            emb = last_hidden[:, -1]
        else:
            sequence_lengths = attention_mask.sum(dim=1) - 1
            batch_size = last_hidden.shape[0]
            emb = last_hidden[torch.arange(batch_size, device=last_hidden.device), sequence_lengths]
    elif pool_type == "cls_last":
        emb = last_hidden[:, 0]
    elif pool_type == "colbert":
        emb = last_hidden
    else:
        raise ValueError(f"pool_type {pool_type} not supported")

    return emb


class LlamaBidirectionalConfig(LlamaConfig):
    """
    Configuration class for LlamaBidirectionalModel.

    Extends LlamaConfig with additional parameters for bidirectional attention
    and pooling configurations.
    """

    model_type = "llama_bidirec"

    def __init__(
        self,
        pooling: str = "avg",
        temperature: float = 1.0,
        **kwargs,
    ):
        """
        Initialize LlamaBidirectionalConfig.

        Args:
            pooling: Pooling strategy ('avg', 'cls', 'last', etc.)
            temperature: Temperature for scaling logits
            **kwargs: Additional arguments passed to LlamaConfig
        """
        self.pooling = pooling
        self.temperature = temperature
        super().__init__(**kwargs)


class LlamaBidirectionalModel(LlamaModel):
    """
    Llama Model with bidirectional attention.

    This model removes causal masking from all attention layers, allowing tokens
    to attend to all other tokens in the sequence. This is useful for embedding
    and retrieval tasks where bidirectional context is beneficial.
    """

    config_class = LlamaBidirectionalConfig

    def __init__(self, config: LlamaConfig):
        """
        Initialize LlamaBidirectionalModel.

        Args:
            config: Model configuration
        """
        super().__init__(config)
        # Disable causal attention for all layers
        for layer in self.layers:
            layer.self_attn.is_causal = False

    def _update_causal_mask(
        self,
        attention_mask: torch.Tensor,
    ):
        if attention_mask is not None and (attention_mask == 0.0).any():
            return attention_mask
        return None

    @check_model_inputs
    @auto_docstring
    def forward(
        self,
        input_ids: Optional[torch.LongTensor] = None,
        attention_mask: Optional[torch.Tensor] = None,
        position_ids: Optional[torch.LongTensor] = None,
        past_key_values: Optional[Cache] = None,
        inputs_embeds: Optional[torch.FloatTensor] = None,
        cache_position: Optional[torch.LongTensor] = None,
        use_cache: Optional[bool] = None,
        **kwargs: Unpack[TransformersKwargs],
    ) -> BaseModelOutputWithPast:
        if (input_ids is None) ^ (inputs_embeds is not None):
            raise ValueError("You must specify exactly one of input_ids or inputs_embeds")

        if inputs_embeds is None:
            inputs_embeds: torch.Tensor = self.embed_tokens(input_ids)

        if use_cache and past_key_values is None:
            past_key_values = DynamicCache(config=self.config)

        if cache_position is None:
            past_seen_tokens = past_key_values.get_seq_length() if past_key_values is not None else 0
            cache_position: torch.Tensor = torch.arange(
                past_seen_tokens, past_seen_tokens + inputs_embeds.shape[1], device=inputs_embeds.device
            )

        if position_ids is None:
            position_ids = cache_position.unsqueeze(0)

        causal_mask = self._update_causal_mask(attention_mask=attention_mask)

        hidden_states = inputs_embeds
        position_embeddings = self.rotary_emb(hidden_states, position_ids)

        for decoder_layer in self.layers[: self.config.num_hidden_layers]:
            hidden_states = decoder_layer(
                hidden_states,
                attention_mask=causal_mask,
                position_ids=position_ids,
                past_key_values=past_key_values,
                cache_position=cache_position,
                position_embeddings=position_embeddings,
                **kwargs,
            )

        hidden_states = self.norm(hidden_states)
        return BaseModelOutputWithPast(
            last_hidden_state=hidden_states,
            past_key_values=past_key_values,
        )


class LlamaBidirectionalForSequenceClassification(LlamaForSequenceClassification):
    """
    Llama Bidirectional Model with a sequence classification/regression head.

    This model adds a classification head on top of the bidirectional Llama model
    and includes configurable pooling strategies.
    """

    config_class = LlamaBidirectionalConfig

    def __init__(self, config):
        """
        Initialize LlamaBidirectionalForSequenceClassification.

        Args:
            config: Model configuration
        """
        super().__init__(config)
        # Release the parameters of LlamaModel
        # created by parent LlamaForSequenceClassification
        del self.model

        self.model = LlamaBidirectionalModel(config)

        # Initialize weights and apply final processing
        self.post_init()

    def forward(
        self,
        input_ids: Optional[torch.LongTensor] = None,
        attention_mask: Optional[torch.Tensor] = None,
        position_ids: Optional[torch.LongTensor] = None,
        past_key_values: Optional[Union[Cache, List[torch.FloatTensor]]] = None,
        inputs_embeds: Optional[torch.FloatTensor] = None,
        labels: Optional[torch.LongTensor] = None,
        use_cache: Optional[bool] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
    ) -> Union[Tuple, SequenceClassifierOutputWithPast]:
        """
        Forward pass for sequence classification.

        Args:
            input_ids: Input token IDs
            attention_mask: Attention mask
            position_ids: Position IDs
            past_key_values: Past key values for generation
            inputs_embeds: Input embeddings (alternative to input_ids)
            labels: Labels for computing loss
            use_cache: Whether to use cache
            output_attentions: Whether to output attentions
            output_hidden_states: Whether to output hidden states
            return_dict: Whether to return a dict

        Returns:
            SequenceClassifierOutputWithPast with loss, logits, and optional outputs
        """
        return_dict = return_dict if return_dict is not None else self.config.use_return_dict

        transformer_outputs = self.model(
            input_ids,
            attention_mask=attention_mask,
            position_ids=position_ids,
            past_key_values=past_key_values,
            inputs_embeds=inputs_embeds,
            use_cache=use_cache,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        )
        hidden_states = transformer_outputs[0]

        # Pool hidden states using configured pooling strategy
        pooled_hidden_states = pool(
            last_hidden_states=hidden_states,
            attention_mask=attention_mask,
            pool_type=self.config.pooling,
        )

        # Apply classification head and temperature scaling
        pooled_logits = self.score(pooled_hidden_states)
        pooled_logits = pooled_logits / self.config.temperature

        loss = None
        if labels is not None:
            labels = labels.to(pooled_logits.device)
            if self.config.problem_type is None:
                if self.num_labels == 1:
                    self.config.problem_type = "regression"
                elif self.num_labels > 1 and (labels.dtype == torch.long or labels.dtype == torch.int):
                    self.config.problem_type = "single_label_classification"
                else:
                    self.config.problem_type = "multi_label_classification"

            if self.config.problem_type == "regression":
                loss_fct = MSELoss()
                if self.num_labels == 1:
                    loss = loss_fct(pooled_logits.squeeze(), labels.squeeze())
                else:
                    loss = loss_fct(pooled_logits, labels)
            elif self.config.problem_type == "single_label_classification":
                loss_fct = CrossEntropyLoss()
                loss = loss_fct(pooled_logits.view(-1, self.num_labels), labels.view(-1))
            elif self.config.problem_type == "multi_label_classification":
                loss_fct = BCEWithLogitsLoss()
                loss = loss_fct(pooled_logits, labels)

        if not return_dict:
            output = (pooled_logits,) + transformer_outputs[1:]
            return ((loss,) + output) if loss is not None else output

        return SequenceClassifierOutputWithPast(
            loss=loss,
            logits=pooled_logits,
            past_key_values=transformer_outputs.past_key_values,
            hidden_states=transformer_outputs.hidden_states,
            attentions=transformer_outputs.attentions,
        )


@dataclass
class BiencoderOutput(ModelOutput):
    """Output dataclass for biencoder model."""

    q_reps: Optional[Tensor] = None
    p_reps: Optional[Tensor] = None
    loss: Optional[Tensor] = None
    labels: Optional[Tensor] = None
    scores: Optional[Tensor] = None


class BiencoderModel(nn.Module):
    """
    Biencoder Model with essential functions for training.

    This model encodes queries and passages separately and computes contrastive loss.
    """

    def __init__(
        self,
        lm_q: PreTrainedModel,
        lm_p: PreTrainedModel,
        linear_pooler: nn.Module = None,
        train_n_passages: int = 1,
        eval_negative_size: int = 0,
        pooling: str = "avg",
        l2_normalize: bool = True,
        t: float = 1.0,
        share_encoder: bool = True,
        add_linear_pooler: bool = False,
    ):
        super().__init__()
        self.lm_q = lm_q
        self.lm_p = lm_p
        self.train_n_passages = train_n_passages
        self.eval_negative_size = eval_negative_size
        self.pooling = pooling
        self.l2_normalize = l2_normalize
        self.t = t
        self.share_encoder = share_encoder
        self.add_linear_pooler = add_linear_pooler
        self.cross_entropy = nn.CrossEntropyLoss(reduction="mean")
        self.linear_pooler = linear_pooler if linear_pooler is not None else nn.Identity()
        self.config = self.lm_q.config
        self.trainer = None

        # For HuggingFace consolidated checkpoint compatibility
        self.name_or_path = os.path.abspath(__file__)
        self.state_dict_adapter = BiencoderStateDictAdapter()
        self.config.architectures = ["LlamaBidirectionalModel"]
        self.config.auto_map = {
            "AutoModel": "llama_bidirectional_model.LlamaBidirectionalModel",
            "AutoConfig": "llama_bidirectional_model.LlamaBidirectionalConfig",
        }

    def forward(self, query: Dict[str, Tensor] = None, passage: Dict[str, Tensor] = None):
        """Forward pass for training."""

        # Get current number of passages per query
        if self.training:
            current_train_n_passages = self.train_n_passages
        else:
            current_train_n_passages = self.eval_negative_size + 1

        # Compute scores (encoding happens inside _compute_scores)
        scores, labels, q_reps, p_reps = self._compute_scores(
            query=query,
            passage=passage,
            current_train_n_passages=current_train_n_passages,
        )
        loss = self.cross_entropy(scores, labels)

        # Adding Dummy Gradients for vlm-based models
        if hasattr(self.lm_q, "module") and hasattr(self.lm_q.module, "post_loss"):
            loss = self.lm_q.module.post_loss(loss, passage)
        elif hasattr(self.lm_q, "post_loss"):
            # Not tested this branch
            loss = self.lm_q.post_loss(loss, passage)

        return BiencoderOutput(
            loss=loss,
            q_reps=q_reps,
            p_reps=p_reps,
            labels=labels.contiguous(),
            scores=scores,
        )

    def _encode(self, encoder: PreTrainedModel, input_dict: dict) -> Optional[torch.Tensor]:
        """Encode input using the encoder."""
        if not input_dict:
            return None

        import inspect

        # Remove token_type_ids if encoder doesn't support it
        if (
            "token_type_ids" not in inspect.getfullargspec(encoder.forward).args
            and "token_type_ids" in input_dict.keys()
        ):
            input_dict = {k: v for k, v in input_dict.items() if k != "token_type_ids"}

        # Get encoder outputs
        outputs = encoder(
            **{k: v for k, v in input_dict.items() if k not in ["kd_labels"]},
            return_dict=True,
            output_hidden_states=True,
        )

        # Extract hidden states
        if hasattr(outputs, "last_hidden_state"):
            hidden_state = outputs.last_hidden_state
        else:
            hidden_state = outputs.hidden_states[-1]

        # Pool the representations
        embeds = pool(
            last_hidden_states=hidden_state,
            attention_mask=input_dict["attention_mask"],
            pool_type=self.pooling,
        )

        # Apply linear pooler
        embeds = self.linear_pooler(embeds)

        # L2 normalize if required
        if self.l2_normalize:
            embeds = F.normalize(embeds, dim=-1)

        return embeds.contiguous()

    def _compute_scores(
        self,
        current_train_n_passages: int,
        query: Dict[str, Tensor] = None,
        passage: Dict[str, Tensor] = None,
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        """Compute similarity scores and labels."""

        # Encode query and passage
        q_reps = self._encode(self.lm_q, query)
        p_reps = self._encode(self.lm_p, passage)

        # Compute similarity scores using contrastive_scores_and_labels
        scores, labels = contrastive_scores_and_labels(
            query=q_reps,
            key=p_reps,
            current_train_n_passages=current_train_n_passages,
        )

        if self.l2_normalize:
            scores = scores / self.t

        return scores, labels, q_reps, p_reps

    @classmethod
    def build(
        cls,
        model_name_or_path: str,
        share_encoder: bool = True,
        add_linear_pooler: bool = False,
        out_dimension: int = 768,
        do_gradient_checkpointing: bool = False,
        train_n_passages: int = 1,
        eval_negative_size: int = 0,
        pooling: str = "avg",
        l2_normalize: bool = True,
        t: float = 1.0,
        trust_remote_code: bool = False,
        **hf_kwargs,
    ):
        """
        Build biencoder model from pretrained.

        Args:
            model_name_or_path: Path to pretrained model or model identifier
            share_encoder: Whether to share encoder weights between query and passage
            add_linear_pooler: Whether to add a linear pooler layer
            out_dimension: Output dimension for linear pooler
            do_gradient_checkpointing: Whether to enable gradient checkpointing
            train_n_passages: Number of passages per query during training
            eval_negative_size: Number of negative samples during evaluation
            pooling: Pooling strategy ('avg', 'cls', 'last', etc.)
            l2_normalize: Whether to L2 normalize embeddings
            t: Temperature for scaling similarity scores
            trust_remote_code: Whether to trust remote code
            **hf_kwargs: Additional arguments passed to model loading
        """

        logger.info(f"Building BiencoderModel from {model_name_or_path}")

        # Infer model class from model_name_or_path
        # Check config.json if it exists
        config_path = os.path.join(model_name_or_path, "config.json")

        if config_path and os.path.exists(config_path):
            import json

            with open(config_path, "r") as f:
                model_type = json.load(f).get("model_type", "")
        else:
            # If no config, infer from model name
            model_type = ""

        # Select model class based on model type from config.json, falling
        # back to the model name/path only when no config was found.
        if "llama" in model_type.lower() or (not model_type and "llama" in model_name_or_path.lower()):
            ModelClass = LlamaBidirectionalModel
            logger.info("Using LlamaBidirectionalModel")
        else:
            raise ValueError(
                f"Unsupported model type: {model_type}. Cannot infer model class from {model_name_or_path}"
            )

        # Load model locally or from hub using selected model class
        if os.path.isdir(model_name_or_path):
            if share_encoder:
                lm_q = ModelClass.from_pretrained(model_name_or_path, trust_remote_code=trust_remote_code, **hf_kwargs)
                lm_p = lm_q
            else:
                _qry_model_path = os.path.join(model_name_or_path, "query_model")
                _psg_model_path = os.path.join(model_name_or_path, "passage_model")

                if not os.path.exists(_qry_model_path):
                    _qry_model_path = model_name_or_path
                    _psg_model_path = model_name_or_path

                lm_q = ModelClass.from_pretrained(_qry_model_path, trust_remote_code=trust_remote_code, **hf_kwargs)
                lm_p = ModelClass.from_pretrained(_psg_model_path, trust_remote_code=trust_remote_code, **hf_kwargs)
        else:
            # Load from hub
            lm_q = ModelClass.from_pretrained(model_name_or_path, **hf_kwargs)

            if share_encoder:
                lm_p = lm_q
            else:
                lm_p = copy.deepcopy(lm_q)

        # Enable gradient checkpointing if requested
        if do_gradient_checkpointing:
            lm_q.gradient_checkpointing_enable()
            if lm_p is not lm_q:
                lm_p.gradient_checkpointing_enable()

        # Create linear pooler if needed
        if add_linear_pooler:
            linear_pooler = nn.Linear(lm_q.config.hidden_size, out_dimension)

            pooler_path = os.path.join(model_name_or_path, "pooler.pt")
            if os.path.exists(pooler_path):
                logger.info("Loading pooler weights from local files")
                state_dict = torch.load(pooler_path, map_location="cpu")
                linear_pooler.load_state_dict(state_dict)
        else:
            linear_pooler = nn.Identity()

        model = cls(
            lm_q=lm_q,
            lm_p=lm_p,
            linear_pooler=linear_pooler,
            train_n_passages=train_n_passages,
            eval_negative_size=eval_negative_size,
            pooling=pooling,
            l2_normalize=l2_normalize,
            t=t,
            share_encoder=share_encoder,
            add_linear_pooler=add_linear_pooler,
        )
        return model

    def save_pretrained(self, save_directory: str, **kwargs):
        """Save model to output directory.

        When a ``checkpointer`` is supplied (the normal training path), saving
        is delegated to :py:meth:`Checkpointer.save_model` so that:

        * The ``state_dict_adapter`` (to_hf / from_hf) conversions are applied,
          keeping the on-disk key format consistent with ``load_model``.
        * Distributed / FSDP state dicts are handled correctly.
        * Async and consolidated-safetensors paths are honoured.

        Without a checkpointer the method falls back to HuggingFace-native
        ``save_pretrained`` on the underlying encoder(s), which is useful for
        standalone / non-distributed export.

        Args:
            save_directory: Output path for the checkpoint.
            **kwargs: Forwarded from the recipe; expected keys include
                ``checkpointer``, ``tokenizer``, and ``peft_config``.
        """
        checkpointer = kwargs.get("checkpointer", None)
        if checkpointer is not None:
            # Delegate to Checkpointer.save_model() which handles:
            # - ModelState.state_dict()
            # - _maybe_adapt_state_dict_to_hf() via state_dict_adapter
            # - Distributed/sharded saving via DCP
            # - Consolidated HF safetensors output
            # - Async checkpointing if enabled
            checkpointer.save_model(
                model=self,
                weights_path=save_directory,
                peft_config=kwargs.get("peft_config", None),
                tokenizer=kwargs.get("tokenizer", None),
            )
            return

        # Fallback: HF-native save (no checkpointer available)
        logger.info(f"Saving BiencoderModel to {save_directory}")

        # Save the model
        if self.share_encoder:
            self.lm_q.save_pretrained(save_directory)
        else:
            os.makedirs(os.path.join(save_directory, "query_model"), exist_ok=True)
            os.makedirs(os.path.join(save_directory, "passage_model"), exist_ok=True)
            self.lm_q.save_pretrained(os.path.join(save_directory, "query_model"))
            self.lm_p.save_pretrained(os.path.join(save_directory, "passage_model"))

        # Save linear pooler if exists
        if self.add_linear_pooler:
            pooler_path = os.path.join(save_directory, "pooler.pt")
            logger.info(f"Saving linear pooler to {pooler_path}")
            torch.save(self.linear_pooler.state_dict(), pooler_path)



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Optional

from torch.distributed.device_mesh import DeviceMesh

from nemo_automodel.components.checkpoint.state_dict_adapter import StateDictAdapter


class BiencoderStateDictAdapter(StateDictAdapter):
    """Adapter for converting BiencoderModel state dict to single encoder format.

    This adapter extracts only the query encoder (lm_q) state dict and converts
    the "lm_q." prefix to "model." prefix, making it compatible with standard
    HuggingFace model format.
    """

    def __init__(self):
        """Initialize the adapter."""
        self._uses_model_prefix = True

    # Prefix added by ModelState for PEFT adapter weights.
    _PEFT_PREFIX = "base_model.model."

    def to_hf(self, state_dict: dict[str, Any], **kwargs) -> dict[str, Any]:
        """Convert from biencoder state dict to HuggingFace format.

        Filters to only lm_q keys and converts "lm_q." prefix to "model." prefix.
        Also handles the ``base_model.model.`` prefix that PEFT checkpointing adds
        so that adapter weights are correctly converted (e.g.
        ``base_model.model.lm_q.X`` → ``base_model.model.model.X``).

        Args:
            state_dict: The biencoder model state dict

        Returns:
            The converted HuggingFace format state dict with only query encoder
        """
        hf_state_dict = {}
        peft_lm_q = self._PEFT_PREFIX + "lm_q."
        peft_pooler = self._PEFT_PREFIX + "linear_pooler."

        for key, value in state_dict.items():
            if key.startswith("lm_q."):
                new_key = "model." + key[len("lm_q.") :]
                hf_state_dict[new_key] = value
            elif key.startswith(peft_lm_q):
                new_key = self._PEFT_PREFIX + "model." + key[len(peft_lm_q) :]
                hf_state_dict[new_key] = value
            elif key.startswith("linear_pooler.") or key.startswith(peft_pooler):
                hf_state_dict[key] = value

        return hf_state_dict

    def from_hf(
        self,
        hf_state_dict: dict[str, Any],
        device_mesh: Optional["DeviceMesh"] = None,
        **kwargs,
    ) -> dict[str, Any]:
        """Convert HuggingFace state dict to biencoder format.

        Converts "model." prefix to "lm_q." prefix for loading into biencoder.
        Also handles the ``base_model.model.`` prefix used by PEFT checkpoints
        (e.g. ``base_model.model.model.X`` → ``base_model.model.lm_q.X``).

        Args:
            hf_state_dict: The HuggingFace format state dict
            device_mesh: Optional device mesh (not used in this adapter)

        Returns:
            The converted biencoder format state dict
        """
        biencoder_state_dict = {}
        peft_model = self._PEFT_PREFIX + "model."
        peft_pooler = self._PEFT_PREFIX + "linear_pooler."

        for key, value in hf_state_dict.items():
            if key.startswith("model."):
                suffix = key[len("model.") :]
                biencoder_state_dict["lm_q." + suffix] = value
                biencoder_state_dict["lm_p." + suffix] = value
            elif key.startswith(peft_model):
                suffix = key[len(peft_model) :]
                biencoder_state_dict[self._PEFT_PREFIX + "lm_q." + suffix] = value
                biencoder_state_dict[self._PEFT_PREFIX + "lm_p." + suffix] = value
            elif key.startswith("linear_pooler.") or key.startswith(peft_pooler):
                biencoder_state_dict[key] = value

        return biencoder_state_dict

    def convert_single_tensor_to_hf(self, fqn: str, tensor: Any, **kwargs) -> list[tuple[str, Any]]:
        """Convert a single tensor from biencoder format to HuggingFace format.

        Args:
            fqn: Fully qualified name of the tensor in biencoder format
            tensor: The tensor to convert
            **kwargs: Additional arguments (unused)

        Returns:
            List of (fqn, tensor) tuples in HuggingFace format.
            Returns empty list if tensor is not part of lm_q.
        """
        if fqn.startswith("lm_q."):
            new_fqn = fqn.replace("lm_q.", "model.")
            return [(new_fqn, tensor)]
        if fqn.startswith("linear_pooler."):
            return [(fqn, tensor)]

        # Skip tensors that are not part of lm_q
        return []



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Adapted from https://github.com/sgl-project/sglang/blob/main/python/sglang/srt/models/registry.py

import importlib
import logging
import pkgutil
from dataclasses import dataclass, field
from functools import lru_cache
from typing import Dict, List, Type, Union

import torch.nn as nn

logger = logging.getLogger(__name__)

MODELING_PATH = ["nemo_automodel.components.models"]


@dataclass
class _ModelRegistry:
    # Keyed by model_arch
    modeling_path: List[str] = field(default_factory=list)
    model_arch_name_to_cls: Dict[str, Union[Type[nn.Module], str]] = field(default_factory=dict)
    naming_override: Dict[str, str] = field(default_factory=dict)

    def __post_init__(self):
        self.naming_override["Qwen3OmniMoeThinkerForConditionalGeneration"] = "Qwen3OmniMoeForConditionalGeneration"
        for modeling_path in self.modeling_path:
            self._mapping_model_arch_name_to_cls(modeling_path)

    @property
    def supported_models(self) -> Dict[str, Type[nn.Module]]:
        return self.model_arch_name_to_cls.keys()

    def get_model_cls_from_model_arch(self, model_arch: str) -> Type[nn.Module]:
        return self.model_arch_name_to_cls[model_arch]

    def register_modeling_path(self, path: str) -> None:
        """Add a new modeling path and register models from it."""
        if path not in self.modeling_path:
            self.modeling_path.append(path)
            self._mapping_model_arch_name_to_cls(path)

    def _mapping_model_arch_name_to_cls(self, modeling_path: str):
        package = importlib.import_module(modeling_path)
        for _, name, ispkg in pkgutil.walk_packages(package.__path__, modeling_path + "."):
            if not ispkg:
                try:
                    module = importlib.import_module(name)
                except Exception as e:
                    logger.warning(f"Ignore import error when loading {name}. {e}")
                    continue
                if hasattr(module, "ModelClass"):
                    entry = module.ModelClass
                    if isinstance(entry, list):
                        for tmp in entry:
                            name = (
                                tmp.__name__
                                if tmp.__name__ not in self.naming_override
                                else self.naming_override[tmp.__name__]
                            )
                            assert name not in self.model_arch_name_to_cls, (
                                f"Duplicated model implementation for {name}"
                            )
                            self.model_arch_name_to_cls[name] = tmp
                    else:
                        name = (
                            entry.__name__
                            if entry.__name__ not in self.naming_override
                            else self.naming_override[entry.__name__]
                        )
                        assert name not in self.model_arch_name_to_cls, f"Duplicated model implementation for {name}"
                        self.model_arch_name_to_cls[name] = entry


@lru_cache
def get_registry():
    return _ModelRegistry(modeling_path=MODELING_PATH)


ModelRegistry = get_registry()



# Copyright (c) 2020, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datasets import load_dataset

from nemo_automodel.components.datasets.utils import SFTSingleTurnPreprocessor


class HellaSwag:
    """A dataset wrapper for the HellaSwag benchmark, tailored for single-turn supervised fine-tuning (SFT).

    This class loads and preprocesses the HellaSwag dataset using a tokenizer and a custom preprocessing
    pipeline for language model fine-tuning. The dataset consists of context and multiple-choice endings,
    where the goal is to choose the most plausible continuation.

    Attributes:
        dataset (Dataset): The processed dataset ready for model training or evaluation.
    """

    def __init__(
        self,
        path_or_dataset,
        tokenizer,
        split="train",
        num_samples_limit=None,
        trust_remote_code=True,
        pad_to_max_length=True,
    ):
        """Initialize the HellaSwag dataset wrapper.

        Args:
            path_or_dataset (str or Dataset): Path to the dataset or a HuggingFace Dataset object.
            tokenizer (PreTrainedTokenizer): The tokenizer used to process text.
            split (str, optional): Dataset split to use (e.g., 'train', 'validation'). Defaults to 'train'.
            num_samples_limit (int, optional): Maximum number of samples to load. Defaults to None.
            trust_remote_code (bool, optional): Whether to trust remote code. Defaults to True.
            pad_to_max_length (bool, optional): Whether to pad sequences to max length in the dataset.
                If False, sequences will have variable lengths and padding will be handled by the collate function.
                Defaults to True.

        Notes:
            If num_samples_limit is an integer, it limits the dataset size using slicing.
        """
        if isinstance(num_samples_limit, int):
            split = f"{split}[:{num_samples_limit}]"
        raw_datasets = load_dataset(path_or_dataset, split=split, trust_remote_code=trust_remote_code)
        processor = SFTSingleTurnPreprocessor(tokenizer)
        processor.pad_to_max_length = pad_to_max_length
        self.dataset = processor.process(raw_datasets, self)

    def get_context(self, examples):
        """Extracts the context part of each example.

        Args:
            examples (dict): A dictionary containing example data with a "ctx" key.

        Returns:
            list[str]: List of context strings.
        """
        return examples["ctx"]

    def get_target(self, examples):
        """Extracts the correct ending based on the label.

        Args:
            examples (dict): A dictionary with "endings" (list of strings) and "label" (index of correct ending).

        Returns:
            list[str]: The gold target strings based on the label index.
        """
        return [endings[int(lbl)] for endings, lbl in zip(examples["endings"], examples["label"], strict=False)]

    def __getitem__(self, index):
        """Get a processed example by index.

        Args:
            index (int): Index of the example.

        Returns:
            dict: A tokenized and preprocessed example.
        """
        ans = self.dataset[index]
        return ans

    def __len__(self):
        """Get the number of examples in the dataset.

        Returns:
            int: Length of the processed dataset.
        """
        return len(self.dataset)



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import enum
import functools
import inspect
from collections.abc import Generator
from contextlib import contextmanager

import yaml


@contextmanager
def safe_yaml_representers() -> Generator[None, None, None]:
    """
    Context manager for safely adding and removing custom YAML representers.

    Temporarily adds custom representers for functions, classes, and other objects
    to the YAML SafeDumper, and restores the original representers when exiting
    the context.

    Usage:
        with safe_yaml_representers():
            yaml_str = yaml.safe_dump(my_complex_object)
    """
    # Save original representers
    original_representers = yaml.SafeDumper.yaml_representers.copy()
    original_multi_representers = yaml.SafeDumper.yaml_multi_representers.copy()

    try:
        # Register custom representers

        # Partial representer
        yaml.SafeDumper.add_representer(functools.partial, _partial_representer)

        # Enum representer
        yaml.SafeDumper.add_multi_representer(enum.Enum, _enum_representer)

        # Function representer
        yaml.SafeDumper.add_representer(type(lambda: ...), _function_representer)
        yaml.SafeDumper.add_representer(type(object), _function_representer)

        # Try to add torch dtype representer if available
        try:
            import torch

            yaml.SafeDumper.add_representer(torch.dtype, _torch_dtype_representer)
        except ModuleNotFoundError:
            pass

        # Try to add GenerationConfig representer if available
        try:
            from transformers import GenerationConfig

            yaml.SafeDumper.add_representer(GenerationConfig, _generation_config_representer)
        except ModuleNotFoundError:
            pass

        # General object representer
        yaml.SafeDumper.add_multi_representer(object, _safe_object_representer)

        yield
    finally:
        # Restore original representers
        yaml.SafeDumper.yaml_representers = original_representers
        yaml.SafeDumper.yaml_multi_representers = original_multi_representers


def _function_representer(dumper, data):
    """Represent functions in YAML."""
    value = {
        "_target_": f"{inspect.getmodule(data).__name__}.{data.__qualname__}",  # type: ignore
        "_call_": False,
    }
    return dumper.represent_data(value)


def _torch_dtype_representer(dumper, data):
    """Represent torch dtypes in YAML."""
    value = {
        "_target_": str(data),
        "_call_": False,
    }
    return dumper.represent_data(value)


def _safe_object_representer(dumper, data):
    """General object representer for YAML.

    This function is a fallback for objects that don't have specific representers.
    If the object has __qualname__ attr,
    the _target_ is set to f"{inspect.getmodule(obj).__name__}.{obj.__qualname__}".
    If the object does not have a __qualname__ attr, the _target_ is set from its __class__ attr.
    The _call_ key is used to indicate whether the target should be called to create an instance.

    Args:
        dumper (yaml.Dumper): The YAML dumper to use for serialization.
        data (Any): The data to serialize.

    Returns:
        The YAML representation of the data.
    """
    try:
        obj = data
        target = f"{inspect.getmodule(obj).__name__}.{obj.__qualname__}"
        call = False
    except AttributeError:
        obj = data.__class__
        target = f"{inspect.getmodule(obj).__name__}.{obj.__qualname__}"
        call = True

    value = {
        "_target_": target,  # type: ignore
        "_call_": call,
    }
    return dumper.represent_data(value)


def _partial_representer(dumper, data):
    """Represent functools.partial objects in YAML."""
    # Get the underlying function
    func = data.func

    # Create a dictionary representation
    value = {
        "_target_": f"{inspect.getmodule(func).__name__}.{func.__qualname__}",
        "_partial_": True,
        "_args_": list(data.args) if data.args else [],
    }

    # Add keyword arguments if any exist
    if data.keywords:
        value |= data.keywords

    return dumper.represent_data(value)


def _enum_representer(dumper, data):
    """Represent enums in YAML."""
    # Create a dictionary representation
    enum_class = data.__class__
    value = {
        "_target_": f"{inspect.getmodule(enum_class).__name__}.{enum_class.__qualname__}",
        "_call_": True,
        "_args_": [data.value],
    }

    return dumper.represent_data(value)


def _generation_config_representer(dumper, data):
    """Represent transformers GenerationConfig objects in YAML."""
    cls = data.__class__
    value = {
        "_target_": f"{inspect.getmodule(cls).__name__}.{cls.__qualname__}.from_dict",
        "_call_": True,
        "config_dict": data.to_dict(),
    }

    return dumper.represent_data(value)



# Copyright (c) 2025, NVIDIA CORPORATION. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
from typing import Any, Dict, Iterable, Optional, Tuple

import torch
import torch.nn as nn

from nemo_automodel.components.distributed.fsdp2 import FSDP2Manager
from nemo_automodel.shared.utils import dtype_from_str

# diffusers is an optional dependency
try:
    from diffusers import DiffusionPipeline

    DIFFUSERS_AVAILABLE = True
except Exception:
    DIFFUSERS_AVAILABLE = False
    DiffusionPipeline = object

logger = logging.getLogger(__name__)


def _choose_device(device: Optional[torch.device]) -> torch.device:
    if device is not None:
        return device
    if torch.cuda.is_available():
        local_rank = int(os.environ.get("LOCAL_RANK", 0))
        return torch.device("cuda", local_rank)
    return torch.device("cpu")


def _iter_pipeline_modules(pipe: DiffusionPipeline) -> Iterable[Tuple[str, nn.Module]]:
    # Prefer Diffusers' components registry when available
    if hasattr(pipe, "components") and isinstance(pipe.components, dict):
        for name, value in pipe.components.items():
            if isinstance(value, nn.Module):
                yield name, value
        return

    # Fallback: inspect attributes
    for name in dir(pipe):
        if name.startswith("_"):
            continue
        try:
            value = getattr(pipe, name)
        except Exception:
            continue
        if isinstance(value, nn.Module):
            yield name, value


def _move_module_to_device(module: nn.Module, device: torch.device, torch_dtype: Any) -> None:
    # torch_dtype can be "auto", torch.dtype, or string
    dtype: Optional[torch.dtype]
    if torch_dtype == "auto":
        dtype = None
    else:
        dtype = dtype_from_str(torch_dtype) if isinstance(torch_dtype, str) else torch_dtype
    if dtype is not None:
        module.to(device=device, dtype=dtype)
    else:
        module.to(device=device)


class NeMoAutoDiffusionPipeline(DiffusionPipeline):
    """
    Drop-in Diffusers pipeline that adds optional FSDP2/TP parallelization during from_pretrained.

    Features:
    - Accepts a per-component mapping from component name to FSDP2Manager
    - Moves all nn.Module components to the chosen device/dtype
    - Parallelizes only components present in the mapping using their manager

    parallel_scheme:
    - Dict[str, FSDP2Manager]: component name -> manager used to parallelize that component
    """

    @classmethod
    def from_pretrained(
        cls,
        pretrained_model_name_or_path: str,
        *model_args,
        parallel_scheme: Optional[Dict[str, FSDP2Manager]] = None,
        device: Optional[torch.device] = None,
        torch_dtype: Any = "auto",
        move_to_device: bool = True,
        **kwargs,
    ) -> "DiffusionPipeline":
        if not DIFFUSERS_AVAILABLE:
            raise RuntimeError(
                "diffusers is required for NeMoAutoDiffusionPipeline.from_pretrained. "
                "Install diffusers with a compatible version."
            )
        pipe = DiffusionPipeline.from_pretrained(
            pretrained_model_name_or_path,
            *model_args,
            torch_dtype=torch_dtype,
            **kwargs,
        )

        # Decide device
        dev = _choose_device(device)

        # Move modules to device/dtype first (helps avoid initial OOM during sharding)
        if move_to_device:
            for name, module in _iter_pipeline_modules(pipe):
                _move_module_to_device(module, dev, torch_dtype)

        # Use per-component FSDP2Manager mappings to parallelize components
        if parallel_scheme is not None:
            assert torch.distributed.is_initialized(), "Expect distributed environment to be initialized"
            for comp_name, comp_module in _iter_pipeline_modules(pipe):
                manager = parallel_scheme.get(comp_name)
                if manager is None:
                    continue
                try:
                    new_m = manager.parallelize(comp_module)
                    if new_m is not comp_module:
                        setattr(pipe, comp_name, new_m)
                except Exception as e:
                    logger.warning("FSDP2Manager.parallelize failed for %s: %s", comp_name, e)
        return pipe



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

from nemo_automodel.components.config.loader import load_yaml_config, translate_value


def parse_cli_argv(default_cfg_path=None):
    """
    Parses CLI args, pulls out --config and collects other --dotted.path options.

    Args:
        default_cfg_path (str, optional): Default config (yaml) path. Defaults to None.

    Raises:
        ValueError: if there's no --config and cfg_path = None
        ValueError: if there's --config but not yaml file passed

    Returns:
        (str, list[str]): the config path along with the config overrides.
    """
    argv = sys.argv[1:]
    overrides = []
    i = 0
    cfg_path = default_cfg_path
    while i < len(argv):
        tok = argv[i]

        # --config or -c
        if tok in ("--config", "-c"):
            if i + 1 >= len(argv):
                raise ValueError("Expected a path after --config")
            cfg_path = argv[i + 1]
            i += 2
            continue

        # any other --option or --dotted.path
        if tok.startswith("--"):
            key = tok.lstrip("-")
            # case A) --key=val
            if "=" in key:
                overrides.append(f"{key}")
                i += 1
            else:
                # case B) --key val
                # if next token exists and isn't another --..., take it as the value
                if i + 1 < len(argv) and not argv[i + 1].startswith("--"):
                    val = argv[i + 1]
                    overrides.append(f"{key}={val}")
                    i += 2
                else:
                    # case C) bare --flag => True
                    overrides.append(f"{key}=True")
                    i += 1
            continue

        # anything else we skip
        i += 1

    if not cfg_path:
        raise ValueError("You must specify --config <path>")
    return cfg_path, overrides


def parse_args_and_load_config(default_cfg_path=None):
    """
    Loads YAML, applies overrides via ConfigNode.set_by_dotted.
    """
    cfg_path, overrides = parse_cli_argv(default_cfg_path)
    print(f"cfg-path: {cfg_path}")
    # load the base YAML
    cfg = load_yaml_config(cfg_path)

    # apply overrides
    for kv in overrides:
        dotted, val_str = kv.split("=", 1)
        cfg.set_by_dotted(dotted, translate_value(val_str))

    return cfg



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import logging
from typing import Dict, List, Optional

from datasets import load_dataset

from nemo_automodel.components.datasets.llm.formatting_utils import _add_pad_token, format_chat_template

logger = logging.getLogger(__name__)

# Map lightweight xLAM types to JSON schema / OpenAI tool types
_TYPE_MAP = {
    "str": "string",
    "string": "string",
    "int": "integer",
    "integer": "integer",
    "float": "number",
    "double": "number",
    "number": "number",
    "bool": "boolean",
    "boolean": "boolean",
    "list": "array",
    "array": "array",
    "object": "object",
}


def _json_load_if_str(value):
    if isinstance(value, str):
        return json.loads(value)
    return value


def _convert_tools(raw_tools: List[Dict]) -> List[Dict]:
    """
    Convert xLAM tool definitions into OpenAI tool schema.
    """
    converted = []
    for tool in raw_tools or []:
        name = tool.get("name")
        if not name:
            logger.warning("Skipping tool without name: %s", tool)
            continue
        description = tool.get("description", "")
        params_raw: Dict = tool.get("parameters") or {}

        properties: Dict[str, Dict] = {}
        required: List[str] = []
        for param_name, param_def in params_raw.items():
            if param_def is None:
                continue
            param_type = str(param_def.get("type", "string")).lower()
            mapped_type = _TYPE_MAP.get(param_type, "string")
            prop = {"type": mapped_type, "description": param_def.get("description", "")}
            if "enum" in param_def:
                prop["enum"] = param_def["enum"]
            if "default" in param_def:
                prop["default"] = param_def["default"]
            else:
                required.append(param_name)
            properties[param_name] = prop

        parameters_schema: Dict = {"type": "object", "properties": properties}
        if required:
            parameters_schema["required"] = required

        converted.append(
            {
                "type": "function",
                "function": {
                    "name": name,
                    "description": description,
                    "parameters": parameters_schema,
                },
            }
        )
    return converted


def _convert_tool_calls(raw_calls: List[Dict], example_id: Optional[int] = None) -> List[Dict]:
    """
    Convert xLAM answers field into OpenAI tool_calls messages.
    """
    tool_calls = []
    for idx, call in enumerate(raw_calls or []):
        name = call.get("name")
        if not name:
            logger.warning("Skipping call without name: %s", call)
            continue
        arguments = call.get("arguments", "")

        call_id = f"call_{example_id}_{idx}" if example_id is not None else f"call_{idx}"
        tool_calls.append(
            {
                "id": call_id,
                "type": "function",
                "function": {
                    "name": name,
                    # Keep arguments as JSON string per OpenAI tool calling format.
                    "arguments": arguments,
                },
            }
        )
    return tool_calls


def _format_example(
    example,
    tokenizer,
    eos_token_id,
    pad_token_id,
    seq_length=None,
    padding=None,
    truncation=None,
):
    tools = _convert_tools(_json_load_if_str(example["tools"]))
    tool_calls = _convert_tool_calls(_json_load_if_str(example["answers"]), example_id=example.get("id"))

    formatted_text = [
        {"role": "user", "content": example["query"]},
        {
            "role": "assistant",
            "content": "",
            "tool_calls": tool_calls,
        },
    ]

    return format_chat_template(
        tokenizer=tokenizer,
        formatted_text=formatted_text,
        tools=tools,
        eos_token_id=eos_token_id,
        pad_token_id=pad_token_id,
        seq_length=seq_length,
        padding=padding,
        truncation=truncation,
        answer_only_loss_mask=True,
    )


def make_xlam_dataset(
    tokenizer,
    seq_length=None,
    limit_dataset_samples=None,
    fp8=False,
    split="train",
    dataset_name="Salesforce/xlam-function-calling-60k",
    padding=False,
    truncation=False,
):
    """
    Load and preprocess the xLAM function-calling dataset to OpenAI messages
    compatible with the bulbasaur chat template (tool-calling aware).

    Each example is formatted as:
      - user: the natural language query
      - assistant: emits tool_calls with serialized arguments
      - tools: OpenAI function schema derived from the dataset tool specs
    """
    if limit_dataset_samples is not None:
        assert isinstance(limit_dataset_samples, int), "Expected limit_dataset_samples to be an int"
        if "[" not in split:
            split = f"{split}[:{limit_dataset_samples}]"
        else:
            logger.warning("Dataset split %s already contains slice, skipping limit_dataset_samples", split)

    dataset = load_dataset(dataset_name, split=split)

    eos_token_id = getattr(tokenizer, "eos_token_id", 0)
    pad_token_id = _add_pad_token(tokenizer) or eos_token_id

    fmt_fn = lambda x: _format_example(  # noqa: E731
        x,
        tokenizer,
        eos_token_id,
        pad_token_id,
        seq_length,
        padding,
        truncation,
    )

    return dataset.map(
        fmt_fn,
        batched=False,
        remove_columns=["id", "query", "answers", "tools"],
    )



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import gc
import math
import re
from typing import Iterable

import torch
from torch.distributed.device_mesh import DeviceMesh
from torch.distributed.tensor import DTensor

from nemo_automodel.components.models.common.utils import set_is_first_microbatch, set_is_optim_step

# Regex pattern to match expert parameters in GroupedExpertsTE.
# Matches FQNs like:
# - model.layers.X.mlp.experts.gate_up_linear.weight0
# - model.layers.X.mlp.experts.gate_up_linear.bias0
# - model.layers.X.mlp.experts.down_linear.weight0
# - model.layers.X.mlp.experts.down_linear.bias0
_TE_EXPERT_PARAM_PATTERN = re.compile(r"(^|\.)mlp\.experts\.(gate_up_linear|down_linear)\.(weight|bias)\d+")


@torch.no_grad()
def count_tail_padding(labels, ignore_label=-100):
    """Counts the total number of padding token in the tail of labels

    e.g.
        labels = torch.tensor([
            [-100, 1, 1, -100, -100],   # 2 tail -100s
            [-100, -100, 2, 3, 4],      # 0 tail -100s
            [5, 6, -100, -100, -100],   # 3 tail -100s
        ])
        count_tail_padding will return 5. Please do note there's more than 5 ignore labels.
    Args:
        labels (torch.Tensor): the labels
        ignore_label (int, optional): ignore label index. Defaults to -100.

    Returns:
        int: total number of ignored tokens in the `labels` input.
    """
    # Flip along the last dimension (seq_len)
    flipped = labels.flip(dims=[1])
    tail_mask = flipped == ignore_label

    # Compute cumulative product to "break" on first non ignore_label
    prod_mask = torch.cumprod(tail_mask.int(), dim=1)

    # Count tail -100s by summing cumprod mask along the sequence dimension
    return prod_mask.view(-1).sum().item()


@torch.no_grad()
def _clip_grad_norm_impl(
    parameters: torch.Tensor | Iterable[torch.Tensor],
    max_norm: float,
    norm_type: float = 2.0,
    error_if_nonfinite: bool = False,
    foreach: bool | None = None,
    pp_mesh: DeviceMesh | None = None,
) -> torch.Tensor:
    # Determine target device for all tensor operations
    # Use current CUDA device if available, otherwise use CPU
    if torch.cuda.is_available():
        target_device = torch.device(f"cuda:{torch.cuda.current_device()}")
    else:
        target_device = torch.device("cpu")

    if isinstance(parameters, torch.Tensor):
        parameters = [parameters]
    else:
        parameters = list(parameters)

    # Group parameters by their sharding pattern
    # Key: (device_mesh_id, tuple of placements)
    sharding_groups = {}

    for p in parameters:
        if p.grad is None:
            continue

        if isinstance(p, DTensor):
            # Create a hashable key from device_mesh and placements
            mesh_id = id(p.device_mesh)
            placements_tuple = tuple(str(placement) for placement in p.placements)
            key = (mesh_id, placements_tuple)
        else:
            # Regular tensor - group separately
            key = ("regular", "regular")

        if key not in sharding_groups:
            sharding_groups[key] = []
        sharding_groups[key].append(p)

    # Compute norm for each sharding group
    group_norms = []
    for group_params in sharding_groups.values():
        grads = [p.grad for p in group_params]
        group_norm = torch.nn.utils.get_total_norm(grads, norm_type, error_if_nonfinite, foreach)
        group_norm = group_norm.float().to(target_device)

        # Convert DTensor norms to regular tensors and ensure they're on the same device
        if isinstance(group_norm, DTensor):
            group_norm = group_norm.full_tensor()

        # Ensure the norm is a regular tensor by cloning and detaching
        # This removes any DTensor metadata that might cause issues
        group_norm = group_norm.clone().detach()

        group_norms.append(group_norm)

    # Combine norms across groups
    if len(group_norms) == 0:
        total_norm = torch.tensor(0.0, device=target_device)
    elif len(group_norms) == 1:
        total_norm = group_norms[0]
    else:
        # Ensure all group norms are on the same device (use the first one's device)
        group_norms = [gn.to(target_device) if gn.device != target_device else gn for gn in group_norms]

        if math.isinf(norm_type):
            # For inf norm, take the maximum across groups
            total_norm = torch.stack(group_norms).max()
        else:
            # For p-norm, combine as (sum of p-th powers)^(1/p)
            total_norm = torch.tensor(0.0, device=target_device)
            for gn in group_norms:
                total_norm += gn**norm_type
            total_norm = total_norm ** (1.0 / norm_type)

    total_norm = total_norm.float().to(target_device)
    # Reduce across pipeline parallel mesh if provided
    if pp_mesh is not None:
        if math.isinf(norm_type):
            torch.distributed.all_reduce(total_norm, op=torch.distributed.ReduceOp.MAX, group=pp_mesh.get_group())
        else:
            total_norm = total_norm**norm_type
            torch.distributed.all_reduce(total_norm, op=torch.distributed.ReduceOp.SUM, group=pp_mesh.get_group())
            total_norm = total_norm ** (1.0 / norm_type)

    # Clip gradients for each sharding group separately
    # This is necessary because clip_grads_with_norm_ doesn't support mixing tensors from different device meshes
    for group_params in sharding_groups.values():
        torch.nn.utils.clip_grads_with_norm_(group_params, max_norm, total_norm, foreach)

    return total_norm


@torch.no_grad()
def clip_grad_norm(
    max_grad_norm: float | None,
    model_parts: list[torch.nn.Module],
    *,
    norm_type: float = 2.0,
    pp_enabled: bool = False,
    device_mesh: DeviceMesh | None = None,
    pp_axis_name: str | None = None,
    foreach: bool = True,
):
    """Common gradient clipping helper.

    Handles all parallelism strategies (TP, PP, EP/MoE) with automatic sharding-aware grouping.
    Returns the gradient norm as a float, or 0.0 if clipping is skipped.

    This function automatically:
    - Groups parameters by sharding pattern (device mesh + placements)
    - Computes norms correctly across different sharding strategies
    - Handles MoE with separate DP/EP meshes
    - Reduces norms across pipeline parallel stages when enabled

    Args:
        max_grad_norm: Maximum gradient norm. If None, skips clipping.
        model_parts: List of model modules to clip.
        norm_type: Type of norm to use (default: 2.0 for L2).
        pp_enabled: Whether pipeline parallelism is enabled.
        device_mesh: Device mesh for parallelism.
        moe_mesh: MoE-specific device mesh (unused, kept for API compatibility).
        ep_axis_name: Expert parallel axis name (unused, kept for API compatibility).
        pp_axis_name: Pipeline parallel axis name.
        foreach: Whether to use foreach implementation for clipping.

    Returns:
        Total gradient norm as a float.
    """
    if max_grad_norm is None:
        return 0.0

    # Collect all parameters
    parameters = [p for m in model_parts for p in m.parameters() if p.requires_grad]

    # Determine pp_mesh if PP is enabled
    pp_mesh = None
    if pp_enabled:
        assert pp_axis_name is not None, "pp_axis_name must be provided when pp_enabled is True"
        pp_mesh = device_mesh[pp_axis_name] if device_mesh is not None else None

    # Use the new sharding-aware implementation
    grad_norm = _clip_grad_norm_impl(
        parameters=parameters,
        max_norm=max_grad_norm,
        norm_type=norm_type,
        error_if_nonfinite=False,
        foreach=foreach,
        pp_mesh=pp_mesh,
    )

    # Convert to float for API compatibility
    if isinstance(grad_norm, torch.Tensor):
        grad_norm = grad_norm.item() if grad_norm.numel() == 1 else grad_norm
        if hasattr(grad_norm, "full_tensor"):
            grad_norm = grad_norm.full_tensor()

    return grad_norm


def prepare_for_grad_accumulation(model_parts: list[torch.nn.Module], pp_enabled: bool = False):
    """Prepare model parts before starting gradient accumulation.

    This is typically called once at the start of gradient accumulation to prepare
    FSDP states for the upcoming forward and backward passes.

    Args:
        model_parts: List of model parts (modules) to prepare.
        pp_enabled: Whether pipeline parallelism is enabled.
    """
    set_is_optim_step(False)
    set_is_first_microbatch(True)
    if pp_enabled:
        return

    for mp in model_parts:
        if hasattr(mp, "prepare_for_grad_accumulation"):
            mp.prepare_for_grad_accumulation(pp_enabled=pp_enabled)


def prepare_after_first_microbatch():
    """Disable first-microbatch flag after the first forward-backward pass.

    Called after the first microbatch in gradient accumulation so that
    subsequent microbatches reuse cached FP8 weights instead of re-quantizing.
    """
    set_is_first_microbatch(False)


def prepare_for_final_backward(model_parts: list[torch.nn.Module], pp_enabled: bool = False):
    """Prepare model parts before the final backward pass.

    This is typically called before the final gradient accumulation step to prepare
    FSDP states for gradient synchronization and resharding.

    Args:
        model_parts: List of model parts (modules) to prepare.
        pp_enabled: Whether pipeline parallelism is enabled.
    """
    set_is_optim_step(True)
    if pp_enabled:
        return

    for mp in model_parts:
        if hasattr(mp, "prepare_for_final_backward"):
            mp.prepare_for_final_backward(pp_enabled=pp_enabled)


@torch.no_grad()
def scale_grads_and_clip_grad_norm(
    max_grad_norm: float | None,
    model_parts: list[torch.nn.Module],
    *,
    norm_type: float = 2.0,
    pp_enabled: bool = False,
    device_mesh: DeviceMesh | None = None,
    moe_mesh: DeviceMesh | None = None,
    ep_axis_name: str | None = None,
    pp_axis_name: str | None = None,
    foreach: bool = True,
    num_label_tokens: int | None = None,
    dp_group_size: int | None = None,
):
    """Scale gradients for PP/EP in a single pass, then clip.

    - PP scaling: divide all local grads by (num_label_tokens / dp_group_size).
    - EP scaling: for parameters on the expert axis, divide grads by (dp_group_size / ep_shard_size).
    - Finally, perform grad clipping with PP/EP-aware reductions.
    """

    # Precompute scale factors
    pp_divisor: float | None = None
    if pp_enabled and num_label_tokens is not None and dp_group_size is not None:
        if dp_group_size != 0:
            candidate = num_label_tokens / dp_group_size
            pp_divisor = float(candidate) if candidate != 0 else None

    ep_ratio: float | None = None
    if moe_mesh is not None and dp_group_size is not None:
        ep_shard_size = moe_mesh["ep_shard"].size() if "ep_shard" in moe_mesh.mesh_dim_names else 1
        if ep_shard_size > 0:
            ep_ratio = float(dp_group_size) / float(ep_shard_size)

    # Single pass over parameters to apply both scalings where applicable
    if pp_divisor is not None or ep_ratio is not None:
        for mp in model_parts:
            for name, p in mp.named_parameters():
                if p.grad is None:
                    continue
                if pp_divisor is not None:
                    p.grad.div_(pp_divisor)
                if ep_ratio is not None:
                    # Scale expert gradients by EP ratio.
                    # DTensor experts: check device mesh for EP sharding axis
                    # Non-DTensor experts (e.g., DeepEP): check param name
                    is_ep_sharded_dtensor = (
                        isinstance(p, DTensor)
                        and isinstance(p.grad, DTensor)
                        and ep_axis_name
                        and ep_axis_name in p.device_mesh.mesh_dim_names
                    )
                    is_expert_param = (
                        isinstance(p, torch.Tensor)
                        and isinstance(p.grad, torch.Tensor)
                        and _TE_EXPERT_PARAM_PATTERN.search(name) is not None
                    )
                    if is_ep_sharded_dtensor or is_expert_param:
                        p.grad.div_(ep_ratio)

    # Clip with the existing PP/EP-aware helper
    return clip_grad_norm(
        max_grad_norm,
        model_parts,
        norm_type=norm_type,
        pp_enabled=pp_enabled,
        device_mesh=device_mesh,
        pp_axis_name=pp_axis_name,
        foreach=foreach,
    )


def move_to_device(model, device):
    # FSDP modules do not move buffers to the device automatically
    for v in model.buffers():
        v.data = v.data.to(device)
    model.to(device)
    gc.collect()
    torch.cuda.empty_cache()


class ScopedModuleOffloading:
    def __init__(self, model, enabled=False):
        self.model = model
        self.enabled = enabled

    def __enter__(self):
        if self.enabled:
            move_to_device(self.model, "cuda")
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.enabled:
            move_to_device(self.model, "cpu")
        return False  # Re-raise exceptions by default



# Copyright (c) 2020, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


MAJOR = 0
MINOR = 3
PATCH = 0
PRE_RELEASE = ""

# Use the following formatting: (major, minor, patch, pre-release)
VERSION = (MAJOR, MINOR, PATCH, PRE_RELEASE)

__shortversion__ = ".".join(map(str, VERSION[:3]))
__version__ = ".".join(map(str, VERSION[:3])) + "".join(VERSION[3:])

__package_name__ = "nemo_automodel"
__contact_names__ = "NVIDIA"
__contact_emails__ = "nemo-toolkit@nvidia.com"
__homepage__ = "https://docs.nvidia.com/deeplearning/nemo/user-guide/docs/en/stable/"
__repository_url__ = "https://github.com/NVIDIA-NeMo/Automodel"
__download_url__ = "https://github.com/NVIDIA-NeMo/Automodel/releases"
__description__ = "NeMo Automodel - Delivers zero-day integration with Hugging Face models, \
                   automating fine-tuning and pretraining with built-in parallelism, \
                   custom-kernels and optimized recipes"
__license__ = "Apache2"
__keywords__ = "deep learning, machine learning, gpu, NLP, NeMo, nvidia, pytorch, torch"



# Copyright (c) 2025, NVIDIA CORPORATION. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# taken and edited from
# https://github.com/pytorch/pytorch/pull/155940
# https://github.com/pytorch/pytorch/pull/155707
# pylint: disable=missing-function-docstring,line-too-long

import dataclasses
import json
import queue
import re
from typing import Any, Optional

import torch
from torch.distributed._shard._utils import narrow_tensor_by_index
from torch.distributed.checkpoint.metadata import (
    ChunkStorageMetadata,
    Metadata,
    MetadataIndex,
    StorageMeta,
    TensorProperties,
    TensorStorageMetadata,
)
from torch.distributed.checkpoint.planner import (
    LoadPlan,
    LoadPlanner,
    ReadItem,
    SavePlan,
    SavePlanner,
    WriteItem,
)
from torch.distributed.checkpoint.storage import WriteResult
from torch.futures import Future

from nemo_automodel.components.checkpoint._backports._fsspec_filesystem import FsspecReader, FsspecWriter
from nemo_automodel.components.checkpoint._backports.consolidate_hf_safetensors import consolidate_safetensors_files
from nemo_automodel.components.checkpoint._backports.filesystem import SerializationFormat
from nemo_automodel.components.checkpoint._backports.hf_utils import (
    CUSTOM_METADATA_KEY,
    DATA_OFFSETS_KEY,
    DEFAULT_EXTRA_METADATA_KEY,
    DTYPE_KEY,
    SAVED_OFFSETS_KEY,
    SHAPE_KEY,
    SUFFIX,
    _gen_file_name,
    _get_dtype,
    _get_safetensors_file_metadata,
    _HFStorageInfo,
    _metadata_fn,
)

__all__ = ["_HuggingFaceStorageWriter", "_HuggingFaceStorageReader"]


class _HuggingFaceStorageWriter(FsspecWriter):
    """
    A writer that writes to a huggingface repository in the huggingface format.
    Uses Fsspec back-end to communicate with back-end storage.
    Fsspec registration of the storage solution is required.
    """

    def __init__(
        self,
        path: str,
        fqn_to_index_mapping: Optional[dict[str, int]] = None,
        thread_count: int = 1,
        token: Optional[str] = None,
        save_sharded: bool = False,
        consolidated_output_path: Optional[str] = None,
        num_threads_consolidation: Optional[int] = None,
        staging_dir: Optional[str] = None,
    ) -> None:
        """
        Initialize the huggingface writer pointing to path.

        Args:
            path: hf directory where the checkpoint will be read from.
                  Needs to have .safetensors files, but can be from any fsspec supported storage,
                  including localFS and hf://.
                  This needs to be a remote path if you want to enable consolidation after saving.
            fqn_to_index_mapping: A mapping from tensor FQN to the index of the file that the tensor should be written to.
                              Indices are from 1 to N, where N is the number of files. If not provided,
                              the tensors will be written to a single file. If none, then all the tensors on the
                              same rank will be written to the same file.
            token: The token to use to authenticate with huggingface hub.
            save_sharded: If True, save the checkpoint as a sharded checkpoint where every rank saves its own shard.
                        Default is False which assumes full tensors are being saved.
            consolidated_output_path: If provided, the output path where the consolidated files will be written in the finish step. This needs to be a local fs path right now.
            num_threads_consolidation: Number of threads to use for parallel processing of saving data to output files. If not provided, the default value is the number of output files.
            staging_dir: Optional directory for staging files during consolidation. If provided,
                        temp files will be created here instead of system temp.
        """
        if token is not None:
            super().__init__(
                path=path,
                token=token,
                serialization_format=SerializationFormat.SAFETENSORS,
            )
        else:
            super().__init__(
                path=path,
                serialization_format=SerializationFormat.SAFETENSORS,
            )
        self._fqn_to_index_mapping: Optional[dict[str, int]] = fqn_to_index_mapping
        self._save_sharded = save_sharded
        self._consolidated_output_path = consolidated_output_path
        self._staging_dir = staging_dir

        if num_threads_consolidation:
            self._num_threads_consolidation = num_threads_consolidation
        elif self._fqn_to_index_mapping:
            self._num_threads_consolidation = max(self._fqn_to_index_mapping.values())
        else:
            self._num_threads_consolidation = 1

        self.thread_count = thread_count

    def prepare_global_plan(self, plans: list[SavePlan]) -> list[SavePlan]:
        new_plans = []
        for i, plan in enumerate(plans, start=1):
            storage_data: dict[str, Any] = {}
            # save default shard mapping. We only use fqn_to_index_mapping for consolidation.
            # if self._fqn_to_index_mapping is not None:
            #     storage_data["fqn_to_index_mapping"] = self._fqn_to_index_mapping
            if self._save_sharded:
                storage_data["shard_index"] = i

            new_plans.append(dataclasses.replace(plan, storage_data=storage_data))

        return new_plans

    def write_data(
        self,
        plan: SavePlan,
        planner: SavePlanner,
    ) -> Future[list[WriteResult]]:
        if len(plan.items) == 0:
            fut: Future = Future()
            fut.set_result([])
            return fut

        # storage_plan is a map from key to file index
        storage_data: dict[str, Any] = plan.storage_data
        storage_plan: Optional[dict[str, int]] = None
        shard_index: Optional[int] = None
        if "fqn_to_index_mapping" in storage_data:
            storage_plan = storage_data["fqn_to_index_mapping"]
        if "shard_index" in storage_data:
            shard_index = storage_data["shard_index"]

        buckets = self._split_by_storage_plan(storage_plan, plan.items)
        highest_index = max(storage_plan.values()) if storage_plan is not None else 1

        file_queue: queue.Queue = queue.Queue()
        for file_index, write_items in buckets.items():
            file_name = _gen_file_name(file_index, highest_index, shard_index)
            file_queue.put((self.fs.concat_path(self.path, file_name), file_name, write_items))

        return super()._write_data(planner, file_queue)

    def finish(self, metadata: Metadata, results: list[list[WriteResult]]) -> None:
        if self._save_sharded and not self._consolidated_output_path:
            return
        if self._save_sharded:
            # Use staging for single-rank consolidation path
            return consolidate_safetensors_files(
                input_dir=self.path,
                output_dir=self._consolidated_output_path,
                num_threads=self._num_threads_consolidation,
                fqn_to_index_mapping=self._fqn_to_index_mapping,
                use_staging=True,
                staging_dir=self._staging_dir,
            )

        metadata_to_write = {}
        storage_md = {}
        total_size = 0
        for wr_list in results:
            storage_md.update({wr.index.fqn: wr.storage_data.relative_path for wr in wr_list})
            total_size += sum([wr.storage_data.length for wr in wr_list])
        metadata_to_write["metadata"] = {"total_size": total_size}
        metadata_to_write["weight_map"] = storage_md

        metadata_path = self.fs.concat_path(self.path, f"{_metadata_fn}")
        with self.fs.create_stream(metadata_path, "w") as metadata_file:
            json.dump(metadata_to_write, metadata_file, indent=2)

    def _split_by_storage_plan(
        self, storage_plan: Optional[dict[str, int]], items: list[WriteItem]
    ) -> dict[int, list[WriteItem]]:
        # storage_plan is a map from key to index
        if storage_plan is None:
            return {1: items}

        buckets = {}
        for item in items:
            key = item.index.fqn

            idx = storage_plan[key]
            if idx not in buckets:
                buckets[idx] = [item]
            else:
                buckets[idx].append(item)

        return buckets

    @property
    def metadata_path(self) -> str:
        return _metadata_fn


class _HuggingFaceStorageReader(FsspecReader):
    """
    A reader that reads from a huggingface repository in the huggingface format.
    Uses in Fsspec back-end to communicate with storage.
    Fsspec registration of the storage solution is required.
    """

    def __init__(self, path: str, token: Optional[str] = None, key_mapping: Optional[dict[str, str]] = None) -> None:
        """
        Initialize the huggingface reader pointing to path.

        Args:
            path: hf directory where the checkpoint will be read from.
            Needs to have .safetensors file, but can be from any fsspec supported storage,
            including localFS and hf://.
            token: The token to use to authenticate with huggingface hub.
            key_mapping: VLMs in HuggingFace can have their FQNs remapped at load time. This means that the state dict keys are not the same as the loaded model's FQNs.
                         This mapping is used to map the state dict keys to the loaded model's FQNs.
        """

        if token is not None:
            super().__init__(path=path, token=token)
        else:
            super().__init__(path=path)

        self.key_mapping = key_mapping

    def read_data(self, plan: LoadPlan, planner: LoadPlanner) -> Future[None]:
        per_file: dict[str, list[ReadItem]] = {}

        for read_item in plan.items:
            item_md: _HFStorageInfo = self.storage_data[read_item.storage_index]
            file_name = item_md.relative_path
            per_file.setdefault(file_name, []).append(read_item)

        for file_name, reqs in per_file.items():
            with self.fs.create_stream(file_name, "rb") as stream:
                for req in reqs:
                    item_md = self.storage_data[req.storage_index]

                    stream.seek(item_md.offset)
                    tensor_bytes = bytearray(stream.read(item_md.length))

                    tensor = torch.frombuffer(
                        tensor_bytes,
                        dtype=item_md.dtype,
                    )
                    tensor = tensor.reshape(item_md.shape)
                    tensor = narrow_tensor_by_index(tensor, req.storage_offsets, req.lengths)
                    target_tensor = planner.resolve_tensor(req).detach()

                    assert target_tensor.size() == tensor.size(), (
                        f"req {req.storage_index} mismatch sizes {target_tensor.size()} vs {tensor.size()}"
                    )

                    target_tensor.copy_(tensor)
                    planner.commit_tensor(req, target_tensor)

        fut: Future = Future()
        fut.set_result(None)
        return fut

    def read_metadata(self) -> Metadata:
        state_dict_metadata: dict[str, TensorStorageMetadata] = {}
        storage_data: dict[MetadataIndex, _HFStorageInfo] = {}

        safetensors_files = []
        for file in self.fs.ls(self.path):
            if file.endswith(SUFFIX):
                safetensors_files.append(file)

        for safetensor_file in safetensors_files:
            with self.fs.create_stream(safetensor_file, "rb") as f:
                safetensors_metadata, metadata_size = _get_safetensors_file_metadata(f)
                custom_metadata = safetensors_metadata.get(DEFAULT_EXTRA_METADATA_KEY)

                dcp_sharding_info = None
                if custom_metadata and custom_metadata.get(CUSTOM_METADATA_KEY):
                    dcp_sharding_info = json.loads(custom_metadata.get(CUSTOM_METADATA_KEY))

                for key, val in safetensors_metadata.items():
                    if key == DEFAULT_EXTRA_METADATA_KEY:
                        continue

                    key = _get_key_renaming_mapping(key, self.key_mapping)

                    # construct state_dict_metadata
                    if dcp_sharding_info is not None:
                        offset = dcp_sharding_info[key][SAVED_OFFSETS_KEY]
                    else:
                        offset = [0] * len(val[SHAPE_KEY])

                    if key not in state_dict_metadata:
                        state_dict_metadata[key] = TensorStorageMetadata(
                            properties=TensorProperties(dtype=_get_dtype(val[DTYPE_KEY])),
                            size=torch.Size([saved + offset for saved, offset in zip(val[SHAPE_KEY], offset)]),
                            chunks=[
                                ChunkStorageMetadata(
                                    offsets=torch.Size(offset),
                                    sizes=torch.Size(val[SHAPE_KEY]),
                                )
                            ],
                        )
                    else:
                        state_dict_metadata[key].chunks.append(
                            ChunkStorageMetadata(torch.Size(offset), sizes=torch.Size(val[SHAPE_KEY]))
                        )
                        size = list(state_dict_metadata[key].size)
                        for i in range(len(size)):
                            size[i] = max(size[i], val[SHAPE_KEY][i] + offset[i])
                        state_dict_metadata[key].size = torch.Size(size)

                    # construct storage data
                    if dcp_sharding_info is not None:
                        metadata_index = MetadataIndex(fqn=key, offset=dcp_sharding_info[key][SAVED_OFFSETS_KEY])
                    else:
                        metadata_index = MetadataIndex(fqn=key, offset=[0] * len(val[SHAPE_KEY]))
                    storage_data[metadata_index] = _HFStorageInfo(
                        relative_path=safetensor_file,
                        offset=val[DATA_OFFSETS_KEY][0] + metadata_size,
                        length=val[DATA_OFFSETS_KEY][1] - val[DATA_OFFSETS_KEY][0],
                        shape=torch.Size(val[SHAPE_KEY]),
                        dtype=_get_dtype(val[DTYPE_KEY]),
                    )

        metadata = Metadata(
            state_dict_metadata=state_dict_metadata,  # type: ignore[arg-type]
            storage_data=storage_data,
        )

        if getattr(metadata, "storage_meta", None) is None:
            metadata.storage_meta = StorageMeta()
        metadata.storage_meta.load_id = self.load_id  # type: ignore[union-attr]

        return metadata


def _extract_file_index(filename: str) -> int:
    """Return the 1-based shard index encoded in a safetensors filename.

    Supported patterns::

        model-00001-of-00008.safetensors
        shard-00000-model-00002-of-00008.safetensors
        model.safetensors  (single-file checkpoints)

    Args:
        filename: The (relative) safetensors filename.

    Returns:
        The numeric shard index, defaulting to ``1`` when no explicit index is
        present or when the filename cannot be parsed.
    """
    # Strip any leading directory components so we only deal with the basename.
    basename = filename.split("/")[-1]

    # Single-file checkpoints which usually carry the name ``model.safetensors``.
    if basename == "model.safetensors":
        return 1

    parts = basename.split("-")

    # Common HF pattern: *model-{idx}-of-{n}.safetensors*
    if "model" in parts:
        idx_pos = parts.index("model") + 1
        if idx_pos < len(parts):
            token = parts[idx_pos].split(".")[0]  # Remove extension if present.
            try:
                return int(token.lstrip("0") or "0")
            except ValueError:
                pass

    # default to the first shard.
    return 1


def get_fqn_to_file_index_mapping(
    reference_model_path: str, key_mapping: Optional[dict[str, str]] = None
) -> dict[str, int]:
    """
    Get the FQN to file index mapping from the metadata.

    Args:
        reference_model_path: Path to reference model to copy file structure from.

    Returns:
        A mapping from tensor FQN to the index of the file that the tensor should be written to.
        Indices are from 1 to N, where N is the number of files.
    """
    hf_reader = _HuggingFaceStorageReader(reference_model_path)
    fqn_to_file_index_mapping: dict[str, int] = {}
    metadata = hf_reader.read_metadata()

    for md_index, storage_info in metadata.storage_data.items():
        fqn = getattr(md_index, "fqn", md_index)
        fqn = _get_key_renaming_mapping(fqn, key_mapping)
        filename = storage_info.relative_path
        fqn_to_file_index_mapping[str(fqn)] = _extract_file_index(filename)

    return fqn_to_file_index_mapping


# the following function is taken from https://github.com/huggingface/transformers/blob/b85ed49e0a5f1bd9fd887f497d055b22b9319a12/src/transformers/modeling_utils.py#L4989-L5047
def _get_key_renaming_mapping(
    key: str,
    key_mapping: Optional[dict[str, str]] = None,
) -> str:
    if key_mapping is None:
        return key

    # Optionally map the key according to `key_mapping`
    for pattern, replacement in key_mapping.items():
        new_key, n_replace = re.subn(pattern, replacement, key)
        # Early exit of the loop
        if n_replace > 0:
            return new_key
    return key
